@@ -22,13 +22,14 @@
 
 #include "system.h"
 
+#include <usual/cfparser.h>
 #include <usual/time.h>
 #include <usual/list.h>
 #include <usual/statlist.h>
 #include <usual/string.h>
 #include <usual/logging.h>
 #include <usual/aatree.h>
-#include <usual/lookup3.h>
+#include <usual/hashing/lookup3.h>
 #include <usual/slab.h>
 #include <usual/socket.h>
 #include <usual/safeio.h>
@@ -42,12 +43,13 @@
 #define FULLVER   PACKAGE_NAME " version " PACKAGE_VERSION
 #endif
 
-/* each state corresponts to a list */
+/* each state corresponds to a list */
 enum SocketState {
 	CL_FREE,		/* free_client_list */
 	CL_JUSTFREE,		/* justfree_client_list */
 	CL_LOGIN,		/* login_client_list */
 	CL_WAITING,		/* pool->waiting_client_list */
+	CL_WAITING_LOGIN,	/*   - but return to CL_LOGIN instead of CL_ACTIVE */
 	CL_ACTIVE,		/* pool->active_client_list */
 	CL_CANCEL,		/* pool->cancel_req_list */
 
@@ -66,6 +68,15 @@ enum PauseMode {
 	P_SUSPEND = 2		/* wait for buffers to be empty */
 };
 
+enum SSLMode {
+	SSLMODE_DISABLED,
+	SSLMODE_ALLOW,
+	SSLMODE_PREFER,
+	SSLMODE_REQUIRE,
+	SSLMODE_VERIFY_CA,
+	SSLMODE_VERIFY_FULL
+};
+
 #define is_server_socket(sk) ((sk)->state >= SV_FREE)
 
 
@@ -97,20 +108,31 @@ extern int cf_sbuf_len;
 #include "stats.h"
 #include "takeover.h"
 #include "janitor.h"
+#include "hba.h"
 
 /* to avoid allocations will use static buffers */
 #define MAX_DBNAME	64
 #define MAX_USERNAME	64
 #define MAX_PASSWORD	64
 
-/* auth modes, should match PG's */
+/* no-auth modes */
 #define AUTH_ANY	-1 /* same as trust but without username check */
-#define AUTH_TRUST	0
+#define AUTH_TRUST	AUTH_OK
+
+/* protocol codes */
+#define AUTH_OK		0
+#define AUTH_KRB	2
 #define AUTH_PLAIN	3
 #define AUTH_CRYPT	4
 #define AUTH_MD5	5
 #define AUTH_CREDS	6
 
+/* internal codes */
+#define AUTH_CERT	7
+#define AUTH_PEER	8
+#define AUTH_HBA	9
+#define AUTH_REJECT	10
+
 /* type codes for weird pkts */
 #define PKT_STARTUP_V2  0x20000
 #define PKT_STARTUP     0x30000
@@ -120,11 +142,7 @@ extern int cf_sbuf_len;
 #define POOL_SESSION	0
 #define POOL_TX		1
 #define POOL_STMT	2
-
-/* old style V2 header: len:4b code:4b */
-#define OLD_HEADER_LEN	8
-/* new style V3 packet header len - type:1b, len:4b */ 
-#define NEW_HEADER_LEN	5
+#define POOL_INHERIT	3
 
 #define BACKENDKEY_LEN	8
 
@@ -139,16 +157,24 @@ extern int cf_sbuf_len;
 /* buffer for pgaddr string conversions (with port) */
 #define PGADDR_BUF  (INET6_ADDRSTRLEN + 10)
 
+struct sockaddr_ucreds {
+	struct sockaddr_in sin;
+	uid_t uid;
+	pid_t pid;
+};
+
 /*
  * AF_INET,AF_INET6 are stored as-is,
- * AF_UNIX uses sockaddr_in port.
+ * AF_UNIX uses sockaddr_in port + uid/pid.
  */
 union PgAddr {
 	struct sockaddr sa;
 	struct sockaddr_in sin;
 	struct sockaddr_in6 sin6;
+	struct sockaddr_ucreds scred;
 };
 
+static inline unsigned int pga_family(const PgAddr *a) { return a->sa.sa_family; }
 static inline bool pga_is_unix(const PgAddr *a) { return a->sa.sa_family == AF_UNIX; }
 
 int pga_port(const PgAddr *a);
@@ -157,6 +183,7 @@ void pga_copy(PgAddr *a, const struct sockaddr *sa);
 bool pga_pton(PgAddr *a, const char *s, int port);
 const char *pga_ntop(const PgAddr *a, char *dst, int dstlen);
 const char *pga_str(const PgAddr *a, char *dst, int dstlen);
+const char *pga_details(const PgAddr *a, char *dst, int dstlen);
 int pga_cmp_addr(const PgAddr *a, const PgAddr *b);
 
 /*
@@ -182,7 +209,7 @@ struct PgPool {
 	struct List head;			/* entry in global pool_list */
 	struct List map_head;			/* entry in user->pool_list */
 
-	PgDatabase *db;			/* corresponging database */
+	PgDatabase *db;			/* corresponding database */
 	PgUser *user;			/* user logged in as */
 
 	struct StatList active_client_list;	/* waiting events logged in clients */
@@ -213,13 +240,16 @@ struct PgPool {
 	unsigned welcome_msg_ready:1;
 };
 
-#define pool_server_count(pool) ( \
+#define pool_connected_server_count(pool) ( \
 		statlist_count(&(pool)->active_server_list) + \
 		statlist_count(&(pool)->idle_server_list) + \
-		statlist_count(&(pool)->new_server_list) + \
 		statlist_count(&(pool)->tested_server_list) + \
 		statlist_count(&(pool)->used_server_list))
 
+#define pool_server_count(pool) ( \
+		pool_connected_server_count(pool) + \
+		statlist_count(&(pool)->new_server_list))
+
 #define pool_client_count(pool) ( \
 		statlist_count(&(pool)->active_client_list) + \
 		statlist_count(&(pool)->waiting_client_list))
@@ -227,13 +257,13 @@ struct PgPool {
 /*
  * A user in login db.
  *
- * fixme: remove ->head as ->tree_node should be enough.
+ * FIXME: remove ->head as ->tree_node should be enough.
  *
  * For databases where remote user is forced, the pool is:
  * first(db->forced_user->pool_list), where pool_list has only one entry.
  *
  * Otherwise, ->pool_list contains multiple pools, for all PgDatabases
- * whis user has logged in.
+ * which user has logged in.
  */
 struct PgUser {
 	struct List head;		/* used to attach user to list */
@@ -241,6 +271,9 @@ struct PgUser {
 	struct AANode tree_node;	/* used to attach user to tree */
 	char name[MAX_USERNAME];
 	char passwd[MAX_PASSWORD];
+	int pool_mode;
+	int max_user_connections;	/* how much server connections are allowed */
+	int connection_count;	/* how much connections are used by user now */
 };
 
 /*
@@ -253,17 +286,21 @@ struct PgDatabase {
 	bool db_paused;		/* PAUSE <db>; was issued */
 	bool db_dead;		/* used on RELOAD/SIGHUP to later detect removed dbs */
 	bool db_auto;		/* is the database auto-created by autodb_connstr */
+	bool db_disabled;	/* is the database accepting new connections? */
 	bool admin;		/* internal console db */
 
 	struct PktBuf *startup_params; /* partial StartupMessage (without user) be sent to server */
 
 	PgUser *forced_user;	/* if not NULL, the user/psw is forced */
+	PgUser *auth_user;	/* if not NULL, users not in userlist.txt will be looked up on the server */
 
 	const char *host;	/* host or unix socket name */
 	int port;
 
 	int pool_size;		/* max server connections in one pool */
 	int res_pool_size;	/* additional server connections in case of trouble */
+	int pool_mode;		/* pool mode for this database */
+	int max_db_connections;	/* max server connections between all pools */
 
 	const char *dbname;	/* server-side name, pointer to inside startup_msg */
 
@@ -272,6 +309,10 @@ struct PgDatabase {
 
 	usec_t inactive_time;	/* when auto-database became inactive (to kill it after timeout) */
 	unsigned active_stamp;	/* set if autodb has connections */
+
+	int connection_count;	/* total connections for this database in all pools */
+
+	struct AATree user_tree;	/* users that have been queried on this database */
 };
 
 
@@ -287,6 +328,8 @@ struct PgSocket {
 
 	PgUser *auth_user;	/* presented login, for client it may differ from pool->user */
 
+	int client_auth_type;	/* auth method decided by hba */
+
 	SocketState state:8;	/* this also specifies socket location */
 
 	bool ready:1;		/* server: accepts new query */
@@ -294,8 +337,12 @@ struct PgSocket {
 	bool close_needed:1;	/* server: this socket must be closed ASAP */
 	bool setting_vars:1;	/* server: setting client vars */
 	bool exec_on_connect:1;	/* server: executing connect_query */
+	bool resetting:1;	/* server: executing reset query from auth login; don't release on flush */
+	bool copy_mode:1;	/* server: in copy stream, ignores any Sync packets */
 
 	bool wait_for_welcome:1;/* client: no server yet in pool, cannot send welcome msg */
+	bool wait_for_user_conn:1;/* client: waiting for auth_conn server connection */
+	bool wait_for_user:1;	/* client: waiting for auth_conn query results */
 
 	bool suspended:1;	/* client/server: if the socket is suspended */
 
@@ -303,6 +350,10 @@ struct PgSocket {
 	bool own_user:1;	/* console client: client with same uid on unix socket */
 	bool wait_for_response:1;/* console client: waits for completion of PAUSE/SUSPEND cmd */
 
+	bool wait_sslchar:1;	/* server: waiting for ssl response: S/N */
+
+	int expect_rfq_count;	/* client: count of ReadyForQuery packets client should see */
+
 	usec_t connect_time;	/* when connection was made */
 	usec_t request_time;	/* last activity time */
 	usec_t query_start;	/* query start moment */
@@ -311,7 +362,10 @@ struct PgSocket {
 	PgAddr remote_addr;	/* ip:port for remote endpoint */
 	PgAddr local_addr;	/* ip:port for local endpoint */
 
-	struct DNSToken *dns_token;	/* ongoing request */
+	union {
+		struct DNSToken *dns_token;	/* ongoing request */
+		PgDatabase *db;			/* cache db while doing auth query */
+	};
 
 	VarCache vars;		/* state of interesting server parameters */
 
@@ -326,7 +380,7 @@ struct PgSocket {
 #define tmp_sk_linkfd	query_start
 /* takeover_clean_socket() needs to clean those up */
 
-/* where the salt is temporarly stored */
+/* where the salt is temporarily stored */
 #define tmp_login_salt  cancel_key
 
 /* main.c */
@@ -348,6 +402,8 @@ extern int cf_default_pool_size;
 extern int cf_min_pool_size;
 extern int cf_res_pool_size;
 extern usec_t cf_res_pool_timeout;
+extern int cf_max_db_connections;
+extern int cf_max_user_connections;
 
 extern char * cf_autodb_connstr;
 extern usec_t cf_autodb_idle_timeout;
@@ -356,6 +412,7 @@ extern usec_t cf_suspend_timeout;
 extern usec_t cf_server_lifetime;
 extern usec_t cf_server_idle_timeout;
 extern char * cf_server_reset_query;
+extern int cf_server_reset_query_always;
 extern char * cf_server_check_query;
 extern usec_t cf_server_check_delay;
 extern usec_t cf_server_connect_timeout;
@@ -368,10 +425,13 @@ extern usec_t cf_idle_transaction_timeout;
 extern int cf_server_round_robin;
 extern int cf_disable_pqexec;
 extern usec_t cf_dns_max_ttl;
+extern usec_t cf_dns_nxdomain_ttl;
 extern usec_t cf_dns_zone_check_period;
 
 extern int cf_auth_type;
 extern char *cf_auth_file;
+extern char *cf_auth_query;
+extern char *cf_auth_hba_file;
 
 extern char *cf_pidfile;
 
@@ -398,10 +458,30 @@ extern int cf_tcp_defer_accept;
 extern int cf_log_connections;
 extern int cf_log_disconnections;
 extern int cf_log_pooler_errors;
+extern int cf_application_name_add_host;
+
+extern int cf_client_tls_sslmode;
+extern char *cf_client_tls_protocols;
+extern char *cf_client_tls_ca_file;
+extern char *cf_client_tls_cert_file;
+extern char *cf_client_tls_key_file;
+extern char *cf_client_tls_ciphers;
+extern char *cf_client_tls_dheparams;
+extern char *cf_client_tls_ecdhecurve;
+
+extern int cf_server_tls_sslmode;
+extern char *cf_server_tls_protocols;
+extern char *cf_server_tls_ca_file;
+extern char *cf_server_tls_cert_file;
+extern char *cf_server_tls_key_file;
+extern char *cf_server_tls_ciphers;
+
+extern const struct CfLookup pool_mode_map[];
 
 extern usec_t g_suspend_start;
 
 extern struct DNSContext *adns;
+extern struct HBA *parsed_hba;
 
 static inline PgSocket * _MUSTCHECK
 pop_socket(struct StatList *slist)
@@ -420,6 +500,14 @@ first_socket(struct StatList *slist)
 	return container_of(slist->head.next, PgSocket, head);
 }
 
+static inline PgSocket *
+last_socket(struct StatList *slist)
+{
+	if (statlist_empty(slist))
+		return NULL;
+	return container_of(slist->head.prev, PgSocket, head);
+}
+
 void load_config(void);
 
 
