@@ -17,6 +17,5 @@
  */
 
 bool client_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *pkt)  _MUSTCHECK;
-bool set_pool(PgSocket *client, const char *dbname, const char *username) _MUSTCHECK;
-
-
+bool set_pool(PgSocket *client, const char *dbname, const char *username, const char *password, bool takeover) _MUSTCHECK;
+bool handle_auth_response(PgSocket *client, PktHdr *pkt);
