@@ -103,44 +103,6 @@ static inline unsigned iobuf_parse_limit(const IOBuf *buf, struct MBuf *mbuf, un
 	return avail;
 }
 
-/* recv */
-static inline int _MUSTCHECK iobuf_recv_limit(IOBuf *io, int fd, unsigned len)
-{
-	uint8_t *pos = io->buf + io->recv_pos;
-	int got;
-	unsigned avail = iobuf_amount_recv(io);
-
-	if (len > avail)
-		len = avail;
-
-	Assert(len > 0);
-
-	got = safe_recv(fd, pos, len, 0);
-	if (got > 0)
-		io->recv_pos += got;
-	return got;
-}
-
-static inline int _MUSTCHECK iobuf_recv_max(IOBuf *io, int fd)
-{
-	return iobuf_recv_limit(io, fd, iobuf_amount_recv(io));
-}
-
-/* send tagged data */
-static inline int _MUSTCHECK iobuf_send_pending(IOBuf *io, int fd)
-{
-	uint8_t *pos = io->buf + io->done_pos;
-	int len, res;
-
-	len = io->parse_pos - io->done_pos;
-	Assert(len > 0);
-
-	res = safe_send(fd, pos, len, 0);
-	if (res > 0)
-		io->done_pos += res;
-	return res;
-}
-
 static inline void iobuf_tag_send(IOBuf *io, unsigned len)
 {
 	Assert(len > 0 && len <= iobuf_amount_parse(io));
