@@ -1,9 +1,9 @@
 /*
  * PgBouncer - Lightweight connection pooler for PostgreSQL.
  * 
- * Copyright (c) 2007 Marko Kreen, Skype Technologies OÜ
+ * Copyright (c) 2007-2009  Marko Kreen, Skype Technologies OÜ
  * 
- * Permission to use, copy, modify, and distribute this software for any
+ * Permission to use, copy, modify, and/or distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
  * copyright notice and this permission notice appear in all copies.
  * 
@@ -20,6 +20,9 @@ void pooler_setup(void);
 bool use_pooler_socket(int fd, bool is_unix) _MUSTCHECK;
 void resume_pooler(void);
 void suspend_pooler(void);
-void get_pooler_fds(int *p_net, int *p_unix);
 void per_loop_pooler_maint(void);
 void pooler_tune_accept(bool on);
+
+typedef bool (*pooler_cb)(void *arg, int fd, const PgAddr *addr);
+bool for_each_pooler_fd(pooler_cb cb, void *arg);
+
