@@ -27,10 +27,11 @@ typedef enum {
 	SBUF_EV_CONNECT_OK,	/* got connection */
 	SBUF_EV_FLUSH,		/* data is sent, buffer empty */
 	SBUF_EV_PKT_CALLBACK,	/* next part of pkt data */
+	SBUF_EV_TLS_READY	/* TLS was established */
 } SBufEvent;
 
 /*
- * If less that this amount of data is pending, then
+ * If less than this amount of data is pending, then
  * prefer to merge it with next recv().
  *
  * It needs to be larger than data handler wants
@@ -39,8 +40,11 @@ typedef enum {
  */
 #define SBUF_SMALL_PKT	64
 
+struct tls;
+
 /* fwd def */
 typedef struct SBuf SBuf;
+typedef struct SBufIO SBufIO;
 
 /* callback should return true if it used one of sbuf_prepare_* on sbuf,
    false if it used sbuf_pause(), sbuf_close() or simply wants to wait for
@@ -52,6 +56,12 @@ typedef bool (*sbuf_cb_t)(SBuf *sbuf,
 /* for some reason, libevent has no typedef for callback */
 typedef void (*sbuf_libevent_cb)(int, short, void *);
 
+struct SBufIO {
+	int (*sbufio_recv)(SBuf *sbuf, void *buf, unsigned int len);
+	int (*sbufio_send)(SBuf *sbuf, const void *data, unsigned int len);
+	int (*sbufio_close)(SBuf *sbuf);
+};
+
 /*
  * Stream Buffer.
  *
@@ -63,6 +73,7 @@ struct SBuf {
 
 	uint8_t wait_type;	/* track wait state */
 	uint8_t pkt_action;	/* method for handling current pkt */
+	uint8_t tls_state;	/* progress of tls */
 
 	int sock;		/* fd for this socket */
 
@@ -73,6 +84,10 @@ struct SBuf {
 	SBuf *dst;		/* target SBuf for current packet */
 
 	IOBuf *io;		/* data buffer, lazily allocated */
+
+	const SBufIO *ops;	/* normal vs. TLS */
+	struct tls *tls;	/* TLS context */
+	const char *tls_host;	/* target hostname */
 };
 
 #define sbuf_socket(sbuf) ((sbuf)->sock)
@@ -81,6 +96,10 @@ void sbuf_init(SBuf *sbuf, sbuf_cb_t proto_fn);
 bool sbuf_accept(SBuf *sbuf, int read_sock, bool is_unix)  _MUSTCHECK;
 bool sbuf_connect(SBuf *sbuf, const struct sockaddr *sa, int sa_len, int timeout_sec)  _MUSTCHECK;
 
+void sbuf_tls_setup(void);
+bool sbuf_tls_accept(SBuf *sbuf)  _MUSTCHECK;
+bool sbuf_tls_connect(SBuf *sbuf, const char *hostname)  _MUSTCHECK;
+
 bool sbuf_pause(SBuf *sbuf) _MUSTCHECK;
 void sbuf_continue(SBuf *sbuf);
 bool sbuf_close(SBuf *sbuf) _MUSTCHECK;
@@ -93,6 +112,7 @@ void sbuf_prepare_fetch(SBuf *sbuf, unsigned amount);
 bool sbuf_answer(SBuf *sbuf, const void *buf, unsigned len)  _MUSTCHECK;
 
 bool sbuf_continue_with_callback(SBuf *sbuf, sbuf_libevent_cb cb)  _MUSTCHECK;
+bool sbuf_use_callback_once(SBuf *sbuf, short ev, sbuf_libevent_cb user_cb) _MUSTCHECK;
 
 /*
  * Returns true if SBuf is has no data buffered
@@ -108,4 +128,22 @@ static inline bool sbuf_is_closed(SBuf *sbuf)
 	return sbuf->sock == 0;
 }
 
+/*
+ * Lowlevel operations.
+ */
+
+static inline int sbuf_op_recv(SBuf *sbuf, void *buf, unsigned int len)
+{
+	return sbuf->ops->sbufio_recv(sbuf, buf, len);
+}
+
+static inline int sbuf_op_send(SBuf *sbuf, const void *buf, unsigned int len)
+{
+	return sbuf->ops->sbufio_send(sbuf, buf, len);
+}
+
+static inline int sbuf_op_close(SBuf *sbuf)
+{
+	return sbuf->ops->sbufio_close(sbuf);
+}
 
