@@ -19,14 +19,24 @@
 #ifndef _USUAL_BASE_WIN32_H_
 #define _USUAL_BASE_WIN32_H_
 
+#include <winsock2.h>
 #include <windows.h>
 
+#include <io.h>
+#include <process.h>
+
+#ifndef ECONNABORTED
 #define ECONNABORTED WSAECONNABORTED
+#endif
+#ifndef EMSGSIZE
 #define EMSGSIZE WSAEMSGSIZE
-#define EINPROGRESS WSAEWOULDBLOCK // WSAEINPROGRESS
+#endif
+#ifndef EINPROGRESS
+#define EINPROGRESS WSAEWOULDBLOCK /* WSAEINPROGRESS */
+#endif
 
 #undef EAGAIN
-#define EAGAIN WSAEWOULDBLOCK // WSAEAGAIN
+#define EAGAIN WSAEWOULDBLOCK /* WSAEAGAIN */
 
 #ifndef EAFNOSUPPORT
 #define EAFNOSUPPORT ENOSYS
@@ -46,6 +56,23 @@
 #define srandom(s) srand(s)
 #define random() rand()
 
+#ifdef _MSC_VER
+
+#define snprintf(fmt, ...) _snprintf(fmt, __VA_ARGS__)
+
+static inline int strcasecmp(const char *a, const char *b)
+{
+	return _stricmp(a, b);
+}
+
+static inline int strncasecmp(const char *a, const char *b, size_t cnt)
+{
+	return _strnicmp(a, b, cnt);
+}
+
+typedef int ssize_t;
+
+#endif
 
 /* getrlimit() */
 #define RLIMIT_NOFILE -1
@@ -82,35 +109,46 @@ struct group {
 static inline struct group *getgrnam(const char *g) { return NULL; }
 static inline struct group *getgrgid(gid_t gid) { return NULL; }
 
-/*
- * Minimal dlopen, dlsym, dlclose, dlerror compat.
- */
+/* format specifiers that should be in <inttypes.h> */
+#ifndef HAVE_INTTYPES_H
 
-#define RTLD_LAZY 1
-#define RTLD_NOW  2
+#define PRId8	"d"
+#define PRId16	"d"
+#define PRId32	"d"
+#define PRId64	"I64d"
 
-static inline void *dlopen(const char *fn, int flag)
-{
-	HMODULE h = LoadLibraryEx(fn, NULL, 0);
-	return h;
-}
+#define PRIi8	"d"
+#define PRIi16	"d"
+#define PRIi32	"d"
+#define PRIi64	"I64d"
 
-static inline void *dlsym(void *hptr, const char *fname)
-{
-	HMODULE h = hptr;
-	FARPROC f = GetProcAddress(h, fname);
-	return f;
-}
+#define PRIo8	"o"
+#define PRIo16	"o"
+#define PRIo32	"o"
+#define PRIo64	"I64o"
 
-static inline int dlclose(void *hptr)
-{
-	HMODULE h = hptr;
-	return FreeLibrary(h) ? 0 : -1;
-}
+#define PRIu8	"u"
+#define PRIu16	"u"
+#define PRIu32	"u"
+#define PRIu64	"I64u"
 
-static inline const char *dlerror(void)
-{
-	return strerror(GetLastError());
-}
+#define PRIx8	"x"
+#define PRIx16	"x"
+#define PRIx32	"x"
+#define PRIx64	"I64x"
+
+#define PRIX8	"X"
+#define PRIX16	"X"
+#define PRIX32	"X"
+#define PRIX64	"I64X"
+
+#endif
+
+#define PRIdZ	"Id"
+#define PRIiZ	"Ii"
+#define PRIoZ	"Io"
+#define PRIuZ	"Iu"
+#define PRIxZ	"Ix"
+#define PRIXZ	"IX"
 
 #endif
