@@ -24,6 +24,7 @@
 
 #include <usual/regex.h>
 #include <usual/netdb.h>
+#include <usual/endian.h>
 
 /* regex elements */
 #define WS0	"[ \t\n\r]*"
@@ -243,7 +244,8 @@ static bool send_one_fd(PgSocket *admin,
 			const char *client_enc,
 			const char *std_strings,
 			const char *datestyle,
-			const char *timezone)
+			const char *timezone,
+			const char *password)
 {
 	struct msghdr msg;
 	struct cmsghdr *cmsg;
@@ -253,9 +255,10 @@ static bool send_one_fd(PgSocket *admin,
 
 	struct PktBuf *pkt = pktbuf_temp();
 
-	pktbuf_write_DataRow(pkt, "issssiqissss",
+	pktbuf_write_DataRow(pkt, "issssiqisssss",
 		      fd, task, user, db, addr, port, ckey, link,
-		      client_enc, std_strings, datestyle, timezone);
+		      client_enc, std_strings, datestyle, timezone,
+		      password);
 	if (pkt->failed)
 		return false;
 	iovec.iov_base = pkt->buf;
@@ -267,7 +270,7 @@ static bool send_one_fd(PgSocket *admin,
 	msg.msg_iovlen = 1;
 
 	/* attach a fd */
-	if (pga_is_unix(&admin->remote_addr) && admin->own_user) {
+	if (pga_is_unix(&admin->remote_addr) && admin->own_user && !admin->sbuf.tls) {
 		msg.msg_control = cntbuf;
 		msg.msg_controllen = sizeof(cntbuf);
 
@@ -282,7 +285,11 @@ static bool send_one_fd(PgSocket *admin,
 
 	slog_debug(admin, "sending socket list: fd=%d, len=%d",
 		   fd, (int)msg.msg_controllen);
-	res = safe_sendmsg(sbuf_socket(&admin->sbuf), &msg, 0);
+	if (msg.msg_controllen) {
+		res = safe_sendmsg(sbuf_socket(&admin->sbuf), &msg, 0);
+	} else {
+		res = sbuf_op_send(&admin->sbuf, pkt->buf, pktbuf_written(pkt));
+	}
 	if (res < 0) {
 		log_error("send_one_fd: sendmsg error: %s", strerror(errno));
 		return false;
@@ -305,11 +312,18 @@ static bool show_one_fd(PgSocket *admin, PgSocket *sk)
 	const struct PStr *datestyle = v->var_list[VDateStyle];
 	const struct PStr *timezone = v->var_list[VTimeZone];
 	char addrbuf[PGADDR_BUF];
+	const char *password = NULL;
+
+	/* Skip TLS sockets */
+	if (sk->sbuf.tls || (sk->link && sk->link->sbuf.tls))
+		return true;
 
 	mbuf_init_fixed_reader(&tmp, sk->cancel_key, 8);
 	if (!mbuf_get_uint64be(&tmp, &ckey))
 		return false;
 
+	if (sk->pool->db->auth_user && sk->auth_user && !find_user(sk->auth_user->name))
+		password = sk->auth_user->passwd;
 
 	return send_one_fd(admin, sbuf_socket(&sk->sbuf),
 			   is_server_socket(sk) ? "server" : "client",
@@ -322,7 +336,8 @@ static bool show_one_fd(PgSocket *admin, PgSocket *sk)
 			   client_encoding ? client_encoding->str : NULL,
 			   std_strings ? std_strings->str : NULL,
 			   datestyle ? datestyle->str : NULL,
-			   timezone ? timezone->str : NULL);
+			   timezone ? timezone->str : NULL,
+			   password);
 }
 
 static bool show_pooler_cb(void *arg, int fd, const PgAddr *a)
@@ -331,7 +346,7 @@ static bool show_pooler_cb(void *arg, int fd, const PgAddr *a)
 
 	return send_one_fd(arg, fd, "pooler", NULL, NULL,
 			   pga_ntop(a, buf, sizeof(buf)), pga_port(a), 0, 0,
-			   NULL, NULL, NULL, NULL);
+			   NULL, NULL, NULL, NULL, NULL);
 }
 
 /* send a row with sendmsg, optionally attaching a fd */
@@ -355,6 +370,19 @@ static bool show_fds_from_list(PgSocket *admin, struct StatList *list)
 	return res;
 }
 
+static PgDatabase *find_or_register_database(PgSocket *admin, const char *name)
+{
+	PgDatabase *db = find_database(name);
+	if (db == NULL) {
+		db = register_auto_database(name);
+		if (db != NULL) {
+			slog_info(admin,
+			          "registered new auto-database: %s", name);
+		}
+	}
+	return db;
+}
+
 /*
  * Command: SHOW FDS
  *
@@ -370,6 +398,7 @@ static bool admin_show_fds(PgSocket *admin, const char *arg)
 	 * Dangerous to show to everybody:
 	 * - can lock pooler as code flips async option
 	 * - show cancel keys for all users
+	 * - shows passwords (md5) for dynamic users
 	 */
 	if (!admin->admin_user)
 		return admin_error(admin, "admin access needed");
@@ -383,13 +412,13 @@ static bool admin_show_fds(PgSocket *admin, const char *arg)
 	/*
 	 * send resultset
 	 */
-	SEND_RowDescription(res, admin, "issssiqissss",
+	SEND_RowDescription(res, admin, "issssiqisssss",
 				 "fd", "task",
 				 "user", "database",
 				 "addr", "port",
 				 "cancel", "link",
 				 "client_encoding", "std_strings",
-				 "datestyle", "timezone");
+				 "datestyle", "timezone", "password");
 	if (res)
 		res = show_pooler_fds(admin);
 
@@ -426,25 +455,36 @@ static bool admin_show_databases(PgSocket *admin, const char *arg)
 	struct List *item;
 	const char *f_user;
 	PktBuf *buf;
+	struct CfValue cv;
+	const char *pool_mode_str;
 
+	cv.extra = pool_mode_map;
 	buf = pktbuf_dynamic(256);
 	if (!buf) {
 		admin_error(admin, "no mem");
 		return true;
 	}
 
-	pktbuf_write_RowDescription(buf, "ssissii",
+	pktbuf_write_RowDescription(buf, "ssissiisii",
 				    "name", "host", "port",
-				    "database", "force_user", "pool_size", "reserve_pool");
+				    "database", "force_user", "pool_size", "reserve_pool",
+				    "pool_mode", "max_connections", "current_connections");
 	statlist_for_each(item, &database_list) {
 		db = container_of(item, PgDatabase, head);
 
 		f_user = db->forced_user ? db->forced_user->name : NULL;
-		pktbuf_write_DataRow(buf, "ssissii",
+		pool_mode_str = NULL;
+		cv.value_p = &db->pool_mode;
+		if (db->pool_mode != POOL_INHERIT)
+			pool_mode_str = cf_get_lookup(&cv);
+		pktbuf_write_DataRow(buf, "ssissiisii",
 				     db->name, db->host, db->port,
 				     db->dbname, f_user,
 				     db->pool_size,
-				     db->res_pool_size);
+				     db->res_pool_size,
+				     pool_mode_str,
+				     database_max_connections(db),
+				     db->connection_count);
 	}
 	admin_flush(admin, buf, "SHOW");
 	return true;
@@ -487,21 +527,31 @@ static bool admin_show_users(PgSocket *admin, const char *arg)
 	PgUser *user;
 	struct List *item;
 	PktBuf *buf = pktbuf_dynamic(256);
+	struct CfValue cv;
+	const char *pool_mode_str;
+
 	if (!buf) {
 		admin_error(admin, "no mem");
 		return true;
 	}
-	pktbuf_write_RowDescription(buf, "s", "name");
+	cv.extra = pool_mode_map;
+
+	pktbuf_write_RowDescription(buf, "ss", "name", "pool_mode");
 	statlist_for_each(item, &user_list) {
 		user = container_of(item, PgUser, head);
-		pktbuf_write_DataRow(buf, "s", user->name);
+		pool_mode_str = NULL;
+		cv.value_p = &user->pool_mode;
+		if (user->pool_mode != POOL_INHERIT)
+			pool_mode_str = cf_get_lookup(&cv);
+
+		pktbuf_write_DataRow(buf, "ss", user->name, pool_mode_str);
 	}
 	admin_flush(admin, buf, "SHOW");
 	return true;
 }
 
-#define SKF_STD "sssssisiTTss"
-#define SKF_DBG "sssssisiTTssiiiiiii"
+#define SKF_STD "sssssisiTTssis"
+#define SKF_DBG "sssssisiTTssisiiiiiii"
 
 static void socket_header(PktBuf *buf, bool debug)
 {
@@ -509,7 +559,8 @@ static void socket_header(PktBuf *buf, bool debug)
 				    "type", "user", "database", "state",
 				    "addr", "port", "local_addr", "local_port",
 				    "connect_time", "request_time",
-				    "ptr", "link",
+				    "ptr", "link", "remote_pid", "tls",
+				    /* debug follows */
 				    "recv_pos", "pkt_pos", "pkt_remain",
 				    "send_pos", "send_remain",
 				    "pkt_avail", "send_avail");
@@ -523,9 +574,11 @@ static void adr2txt(const PgAddr *adr, char *dst, unsigned dstlen)
 static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
 {
 	int pkt_avail = 0, send_avail = 0;
+	int remote_pid;
 	char ptrbuf[128], linkbuf[128];
 	char l_addr[PGADDR_BUF], r_addr[PGADDR_BUF];
 	IOBuf *io = sk->sbuf.io;
+	char infobuf[96] = "";
 
 	if (io) {
 		pkt_avail = iobuf_amount_parse(sk->sbuf.io);
@@ -541,6 +594,18 @@ static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
 	else
 		linkbuf[0] = 0;
 
+	/* get pid over unix socket */
+	if (pga_is_unix(&sk->remote_addr))
+		remote_pid = sk->remote_addr.scred.pid;
+	else
+		remote_pid = 0;
+	/* if that failed, get it from cancel key */
+	if (is_server_socket(sk) && remote_pid == 0)
+		remote_pid = be32dec(sk->cancel_key);
+
+	if (sk->sbuf.tls)
+		tls_get_connection_info(sk->sbuf.tls, infobuf, sizeof infobuf);
+
 	pktbuf_write_DataRow(buf, debug ? SKF_DBG : SKF_STD,
 			     is_server_socket(sk) ? "S" :"C",
 			     sk->auth_user ? sk->auth_user->name : "(nouser)",
@@ -549,7 +614,8 @@ static void socket_row(PktBuf *buf, PgSocket *sk, const char *state, bool debug)
 			     l_addr, pga_port(&sk->local_addr),
 			     sk->connect_time,
 			     sk->request_time,
-			     ptrbuf, linkbuf,
+			     ptrbuf, linkbuf, remote_pid, infobuf,
+			     /* debug */
 			     io ? io->recv_pos : 0,
 			     io ? io->parse_pos : 0,
 			     sk->sbuf.pkt_remain,
@@ -698,22 +764,28 @@ static bool admin_show_pools(PgSocket *admin, const char *arg)
 	PktBuf *buf;
 	PgSocket *waiter;
 	usec_t now = get_cached_time();
+	struct CfValue cv;
+	int pool_mode;
 
+	cv.extra = pool_mode_map;
+	cv.value_p = &pool_mode;
 	buf = pktbuf_dynamic(256);
 	if (!buf) {
 		admin_error(admin, "no mem");
 		return true;
 	}
-	pktbuf_write_RowDescription(buf, "ssiiiiiiii",
+	pktbuf_write_RowDescription(buf, "ssiiiiiiiis",
 				    "database", "user",
 				    "cl_active", "cl_waiting",
 				    "sv_active", "sv_idle",
 				    "sv_used", "sv_tested",
-				    "sv_login", "maxwait");
+				    "sv_login", "maxwait",
+				    "pool_mode");
 	statlist_for_each(item, &pool_list) {
 		pool = container_of(item, PgPool, head);
 		waiter = first_socket(&pool->waiting_client_list);
-		pktbuf_write_DataRow(buf, "ssiiiiiiii",
+		pool_mode = pool_pool_mode(pool);
+		pktbuf_write_DataRow(buf, "ssiiiiiiiis",
 				     pool->db->name, pool->user->name,
 				     statlist_count(&pool->active_client_list),
 				     statlist_count(&pool->waiting_client_list),
@@ -724,7 +796,8 @@ static bool admin_show_pools(PgSocket *admin, const char *arg)
 				     statlist_count(&pool->new_server_list),
 				     /* how long is the oldest client waited */
 				     (waiter && waiter->query_start)
-				     ?  (int)((now - waiter->query_start) / USEC) : 0);
+				     ?  (int)((now - waiter->query_start) / USEC) : 0,
+				     cf_get_lookup(&cv));
 	}
 	admin_flush(admin, buf, "SHOW");
 	return true;
@@ -869,7 +942,7 @@ static bool admin_cmd_shutdown(PgSocket *admin, const char *arg)
 
 	/*
 	 * note: new pooler expects unix socket file gone when it gets
-	 * event from fd.  currently atexit() cleanup should be called
+	 * event from fd.  Currently atexit() cleanup should be called
 	 * before closing open sockets.
 	 */
 	log_info("SHUTDOWN command issued");
@@ -959,15 +1032,9 @@ static bool admin_cmd_pause(PgSocket *admin, const char *arg)
 	} else {
 		PgDatabase *db;
 		log_info("PAUSE '%s' command issued", arg);
-		db = find_database(arg);
-		if (db == NULL) {
-			db = register_auto_database(arg);
-			if (db == NULL) {
-				return admin_error(admin, "no such database: %s", arg);
-			} else {
-				slog_info(admin, "registered new auto-database for PAUSE: %s", arg);
-			}
-		}
+		db = find_or_register_database(admin, arg);
+		if (db == NULL)
+			return admin_error(admin, "no such database: %s", arg);
 		if (db == admin->pool->db)
 			return admin_error(admin, "cannot pause admin db: %s", arg);
 		db->db_paused = 1;
@@ -980,6 +1047,50 @@ static bool admin_cmd_pause(PgSocket *admin, const char *arg)
 	return true;
 }
 
+/* Command: DISABLE */
+static bool admin_cmd_disable(PgSocket *admin, const char *arg)
+{
+	PgDatabase *db;
+
+	if (!admin->admin_user)
+		return admin_error(admin, "admin access needed");
+
+	if (!arg[0])
+		return admin_error(admin, "a database is required");
+
+	log_info("DISABLE '%s' command issued", arg);
+	db = find_or_register_database(admin, arg);
+	if (db == NULL)
+		return admin_error(admin, "no such database: %s", arg);
+	if (db->admin)
+		return admin_error(admin, "cannot disable admin db: %s", arg);
+
+	db->db_disabled = 1;
+	return admin_ready(admin, "DISABLE");
+}
+
+/* Command: ENABLE */
+static bool admin_cmd_enable(PgSocket *admin, const char *arg)
+{
+	PgDatabase *db;
+
+	if (!admin->admin_user)
+		return admin_error(admin, "admin access needed");
+
+	if (!arg[0])
+		return admin_error(admin, "a database is required");
+
+	log_info("ENABLE '%s' command issued", arg);
+	db = find_database(arg);
+	if (db == NULL)
+		return admin_error(admin, "no such database: %s", arg);
+	if (db->admin)
+		return admin_error(admin, "cannot disable admin db: %s", arg);
+
+	db->db_disabled = 0;
+	return admin_ready(admin, "ENABLE");
+}
+
 /* Command: KILL */
 static bool admin_cmd_kill(PgSocket *admin, const char *arg)
 {
@@ -997,15 +1108,9 @@ static bool admin_cmd_kill(PgSocket *admin, const char *arg)
 		return admin_error(admin, "a database is required");
 
 	log_info("KILL '%s' command issued", arg);
-	db = find_database(arg);
-	if (db == NULL) {
-		db = register_auto_database(arg);
-		if (db == NULL) {
-			return admin_error(admin, "no such database: %s", arg);
-		} else {
-			slog_info(admin, "registered new auto-database for KILL: %s", arg);
-		}
-	}
+	db = find_or_register_database(admin, arg);
+	if (db == NULL)
+		return admin_error(admin, "no such database: %s", arg);
 	if (db == admin->pool->db)
 		return admin_error(admin, "cannot kill admin db: %s", arg);
 
@@ -1074,6 +1179,8 @@ static bool admin_show_help(PgSocket *admin, const char *arg)
 		"\tRELOAD\n"
 		"\tPAUSE [<db>]\n"
 		"\tRESUME [<db>]\n"
+		"\tDISABLE <db>\n"
+		"\tENABLE <db>\n"
 		"\tKILL <db>\n"
 		"\tSUSPEND\n"
 		"\tSHUTDOWN", "");
@@ -1133,6 +1240,8 @@ static bool admin_cmd_show(PgSocket *admin, const char *arg)
 }
 
 static struct cmd_lookup cmd_list [] = {
+	{"disable", admin_cmd_disable},
+	{"enable", admin_cmd_enable},
 	{"kill", admin_cmd_kill},
 	{"pause", admin_cmd_pause},
 	{"reload", admin_cmd_reload},
@@ -1198,7 +1307,7 @@ bool admin_handle_client(PgSocket *admin, PktHdr *pkt)
 	const char *q;
 	bool res;
 
-	/* dont tolerate partial packets */
+	/* don't tolerate partial packets */
 	if (incomplete_pkt(pkt)) {
 		disconnect_client(admin, true, "incomplete pkt");
 		return false;
@@ -1230,12 +1339,11 @@ bool admin_handle_client(PgSocket *admin, PktHdr *pkt)
  * Client is unauthenticated, look if it wants to connect
  * to special "pgbouncer" user.
  */
-bool admin_pre_login(PgSocket *client)
+bool admin_pre_login(PgSocket *client, const char *username)
 {
 	uid_t peer_uid = -1;
 	gid_t peer_gid = -1;
 	int res;
-	const char *username = client->auth_user->name;
 
 	client->admin_user = 0;
 	client->own_user = 0;
@@ -1246,6 +1354,7 @@ bool admin_pre_login(PgSocket *client)
 		if (res >= 0 && peer_uid == getuid()
 			&& strcmp("pgbouncer", username) == 0)
 		{
+			client->auth_user = admin_pool->db->forced_user;
 			client->own_user = 1;
 			client->admin_user = 1;
 			slog_info(client, "pgbouncer access from unix socket");
@@ -1255,21 +1364,35 @@ bool admin_pre_login(PgSocket *client)
 
 	/*
 	 * auth_mode=any does not keep original username around,
-	 * so username based checks do not work.
+	 * so username based check has to take place here
 	 */
 	if (cf_auth_type == AUTH_ANY) {
-		if (cf_log_connections)
-			slog_info(client, "auth_mode=any: allowing anybody in as admin");
-		client->admin_user = 1;
-		return true;
+		if (strlist_contains(cf_admin_users, username)) {
+			client->auth_user = admin_pool->db->forced_user;
+			client->admin_user = 1;
+			return true;
+		} else if (strlist_contains(cf_stats_users, username)) {
+			client->auth_user = admin_pool->db->forced_user;
+			return true;
+		}
 	}
+	return false;
+}
 
-	if (strlist_contains(cf_admin_users, username)) {
+bool admin_post_login(PgSocket *client)
+{
+	const char *username = client->auth_user->name;
+
+	if (cf_auth_type == AUTH_ANY)
+		return true;
+
+	if (client->admin_user || strlist_contains(cf_admin_users, username)) {
 		client->admin_user = 1;
 		return true;
 	} else if (strlist_contains(cf_stats_users, username)) {
 		return true;
 	}
+
 	disconnect_client(client, true, "not allowed");
 	return false;
 }
@@ -1291,6 +1414,7 @@ void admin_setup(void)
 	db->port = cf_listen_port;
 	db->pool_size = 2;
 	db->admin = 1;
+	db->pool_mode = POOL_STMT;
 	if (!force_user(db, "pgbouncer", ""))
 		fatal("no mem on startup - cannot alloc pgbouncer user");
 
@@ -1300,10 +1424,14 @@ void admin_setup(void)
 		fatal("cannot create admin pool?");
 	admin_pool = pool;
 
-	/* fake user, with disabled psw */
-	user = add_user("pgbouncer", "");
-	if (!user)
-		fatal("cannot create admin user?");
+	/* user */
+	user = find_user("pgbouncer");
+	if (!user) {
+		/* fake user with disabled psw */
+		user = add_user("pgbouncer", "");
+		if (!user)
+			fatal("cannot create admin user?");
+	}
 
 	/* prepare welcome */
 	msg = pktbuf_dynamic(128);
@@ -1367,7 +1495,7 @@ void admin_pause_done(void)
 			if (count_paused_databases() > 0)
 				res = admin_ready(admin, "PAUSE");
 			else
-				/* fixme */
+				/* FIXME */
 				fatal("admin_pause_done: bad state");
 		}
 
