@@ -17,7 +17,7 @@
  */
 
 /*
- * Launcer for all the rest.
+ * Launcher for all the rest.
  */
 
 #include "bouncer.h"
@@ -50,6 +50,8 @@ static void usage(int err, char *exe)
 /* async dns handler */
 struct DNSContext *adns;
 
+struct HBA *parsed_hba;
+
 /*
  * configuration storage
  */
@@ -86,19 +88,25 @@ int cf_tcp_keepintvl;
 
 int cf_auth_type = AUTH_MD5;
 char *cf_auth_file;
+char *cf_auth_hba_file;
+char *cf_auth_query;
 
 int cf_max_client_conn;
 int cf_default_pool_size;
 int cf_min_pool_size;
 int cf_res_pool_size;
 usec_t cf_res_pool_timeout;
+int cf_max_db_connections;
+int cf_max_user_connections;
 
 char *cf_server_reset_query;
+int cf_server_reset_query_always;
 char *cf_server_check_query;
 usec_t cf_server_check_delay;
 int cf_server_round_robin;
 int cf_disable_pqexec;
 usec_t cf_dns_max_ttl;
+usec_t cf_dns_nxdomain_ttl;
 usec_t cf_dns_zone_check_period;
 unsigned int cf_max_packet_size;
 
@@ -131,6 +139,23 @@ int cf_stats_period;
 int cf_log_connections;
 int cf_log_disconnections;
 int cf_log_pooler_errors;
+int cf_application_name_add_host;
+
+int cf_client_tls_sslmode;
+char *cf_client_tls_protocols;
+char *cf_client_tls_ca_file;
+char *cf_client_tls_cert_file;
+char *cf_client_tls_key_file;
+char *cf_client_tls_ciphers;
+char *cf_client_tls_dheparams;
+char *cf_client_tls_ecdhecurve;
+
+int cf_server_tls_sslmode;
+char *cf_server_tls_protocols;
+char *cf_server_tls_ca_file;
+char *cf_server_tls_cert_file;
+char *cf_server_tls_key_file;
+char *cf_server_tls_ciphers;
 
 /*
  * config file description
@@ -143,20 +168,29 @@ static const struct CfLookup auth_type_map[] = {
 	{ "any", AUTH_ANY },
 	{ "trust", AUTH_TRUST },
 	{ "plain", AUTH_PLAIN },
-#ifdef HAVE_CRYPT
-	{ "crypt", AUTH_CRYPT },
-#endif
 	{ "md5", AUTH_MD5 },
+	{ "cert", AUTH_CERT },
+	{ "hba", AUTH_HBA },
 	{ NULL }
 };
 
-static const struct CfLookup pool_mode_map[] = {
+const struct CfLookup pool_mode_map[] = {
 	{ "session", POOL_SESSION },
 	{ "transaction", POOL_TX },
 	{ "statement", POOL_STMT },
 	{ NULL }
 };
 
+const struct CfLookup sslmode_map[] = {
+	{ "disabled", SSLMODE_DISABLED },
+	{ "allow", SSLMODE_ALLOW },
+	{ "prefer", SSLMODE_PREFER },
+	{ "require", SSLMODE_REQUIRE },
+	{ "verify-ca", SSLMODE_VERIFY_CA },
+	{ "verify-full", SSLMODE_VERIFY_FULL },
+	{ NULL }
+};
+
 static const struct CfKey bouncer_params [] = {
 CF_ABS("job_name", CF_STR, cf_jobname, CF_NO_RELOAD, "pgbouncer"),
 #ifdef WIN32
@@ -175,12 +209,16 @@ CF_ABS("unix_socket_group", CF_STR, cf_unix_socket_group, CF_NO_RELOAD, ""),
 #endif
 CF_ABS("auth_type", CF_LOOKUP(auth_type_map), cf_auth_type, 0, "md5"),
 CF_ABS("auth_file", CF_STR, cf_auth_file, 0, "unconfigured_file"),
+CF_ABS("auth_hba_file", CF_STR, cf_auth_hba_file, 0, ""),
+CF_ABS("auth_query", CF_STR, cf_auth_query, 0, "SELECT usename, passwd FROM pg_shadow WHERE usename=$1"),
 CF_ABS("pool_mode", CF_LOOKUP(pool_mode_map), cf_pool_mode, 0, "session"),
 CF_ABS("max_client_conn", CF_INT, cf_max_client_conn, 0, "100"),
 CF_ABS("default_pool_size", CF_INT, cf_default_pool_size, 0, "20"),
 CF_ABS("min_pool_size", CF_INT, cf_min_pool_size, 0, "0"),
 CF_ABS("reserve_pool_size", CF_INT, cf_res_pool_size, 0, "0"),
 CF_ABS("reserve_pool_timeout", CF_TIME_USEC, cf_res_pool_timeout, 0, "5"),
+CF_ABS("max_db_connections", CF_INT, cf_max_db_connections, 0, "0"),
+CF_ABS("max_user_connections", CF_INT, cf_max_user_connections, 0, "0"),
 CF_ABS("syslog", CF_INT, cf_syslog, 0, "0"),
 CF_ABS("syslog_facility", CF_STR, cf_syslog_facility, 0, "daemon"),
 CF_ABS("syslog_ident", CF_STR, cf_syslog_ident, 0, "pgbouncer"),
@@ -191,10 +229,11 @@ CF_ABS("user", CF_STR, cf_username, CF_NO_RELOAD, NULL),
 CF_ABS("autodb_idle_timeout", CF_TIME_USEC, cf_autodb_idle_timeout, 0, "3600"),
 
 CF_ABS("server_reset_query", CF_STR, cf_server_reset_query, 0, "DISCARD ALL"),
+CF_ABS("server_reset_query_always", CF_INT, cf_server_reset_query_always, 0, "0"),
 CF_ABS("server_check_query", CF_STR, cf_server_check_query, 0, "select 1"),
 CF_ABS("server_check_delay", CF_TIME_USEC, cf_server_check_delay, 0, "30"),
 CF_ABS("query_timeout", CF_TIME_USEC, cf_query_timeout, 0, "0"),
-CF_ABS("query_wait_timeout", CF_TIME_USEC, cf_query_wait_timeout, 0, "0"),
+CF_ABS("query_wait_timeout", CF_TIME_USEC, cf_query_wait_timeout, 0, "120"),
 CF_ABS("client_idle_timeout", CF_TIME_USEC, cf_client_idle_timeout, 0, "0"),
 CF_ABS("client_login_timeout", CF_TIME_USEC, cf_client_login_timeout, 0, "60"),
 CF_ABS("idle_transaction_timeout", CF_TIME_USEC, cf_idle_transaction_timeout, 0, "0"),
@@ -207,10 +246,11 @@ CF_ABS("suspend_timeout", CF_TIME_USEC, cf_suspend_timeout, 0, "10"),
 CF_ABS("ignore_startup_parameters", CF_STR, cf_ignore_startup_params, 0, ""),
 CF_ABS("disable_pqexec", CF_INT, cf_disable_pqexec, CF_NO_RELOAD, "0"),
 CF_ABS("dns_max_ttl", CF_TIME_USEC, cf_dns_max_ttl, 0, "15"),
+CF_ABS("dns_nxdomain_ttl", CF_TIME_USEC, cf_dns_nxdomain_ttl, 0, "15"),
 CF_ABS("dns_zone_check_period", CF_TIME_USEC, cf_dns_zone_check_period, 0, "0"),
 
 CF_ABS("max_packet_size", CF_UINT, cf_max_packet_size, 0, "2147483647"),
-CF_ABS("pkt_buf", CF_INT, cf_sbuf_len, CF_NO_RELOAD, "2048"),
+CF_ABS("pkt_buf", CF_INT, cf_sbuf_len, CF_NO_RELOAD, "4096"),
 CF_ABS("sbuf_loopcnt", CF_INT, cf_sbuf_loopcnt, 0, "5"),
 CF_ABS("tcp_defer_accept", DEFER_OPS, cf_tcp_defer_accept, 0, NULL),
 CF_ABS("tcp_socket_buffer", CF_INT, cf_tcp_socket_buffer, 0, "0"),
@@ -225,6 +265,24 @@ CF_ABS("stats_period", CF_INT, cf_stats_period, 0, "60"),
 CF_ABS("log_connections", CF_INT, cf_log_connections, 0, "1"),
 CF_ABS("log_disconnections", CF_INT, cf_log_disconnections, 0, "1"),
 CF_ABS("log_pooler_errors", CF_INT, cf_log_pooler_errors, 0, "1"),
+CF_ABS("application_name_add_host", CF_INT, cf_application_name_add_host, 0, "0"),
+
+CF_ABS("client_tls_sslmode", CF_LOOKUP(sslmode_map), cf_client_tls_sslmode, CF_NO_RELOAD, "disabled"),
+CF_ABS("client_tls_ca_file", CF_STR, cf_client_tls_ca_file, CF_NO_RELOAD, ""),
+CF_ABS("client_tls_cert_file", CF_STR, cf_client_tls_cert_file, CF_NO_RELOAD, ""),
+CF_ABS("client_tls_key_file", CF_STR, cf_client_tls_key_file, CF_NO_RELOAD, ""),
+CF_ABS("client_tls_protocols", CF_STR, cf_client_tls_protocols, CF_NO_RELOAD, "all"),
+CF_ABS("client_tls_ciphers", CF_STR, cf_client_tls_ciphers, CF_NO_RELOAD, "fast"),
+CF_ABS("client_tls_dheparams", CF_STR, cf_client_tls_dheparams, CF_NO_RELOAD, "auto"),
+CF_ABS("client_tls_ecdhcurve", CF_STR, cf_client_tls_ecdhecurve, CF_NO_RELOAD, "auto"),
+
+CF_ABS("server_tls_sslmode", CF_LOOKUP(sslmode_map), cf_server_tls_sslmode, CF_NO_RELOAD, "disabled"),
+CF_ABS("server_tls_ca_file", CF_STR, cf_server_tls_ca_file, CF_NO_RELOAD, ""),
+CF_ABS("server_tls_cert_file", CF_STR, cf_server_tls_cert_file, CF_NO_RELOAD, ""),
+CF_ABS("server_tls_key_file", CF_STR, cf_server_tls_key_file, CF_NO_RELOAD, ""),
+CF_ABS("server_tls_protocols", CF_STR, cf_server_tls_protocols, CF_NO_RELOAD, "all"),
+CF_ABS("server_tls_ciphers", CF_STR, cf_server_tls_ciphers, CF_NO_RELOAD, "fast"),
+
 {NULL}
 };
 
@@ -235,6 +293,9 @@ static const struct CfSect config_sects [] = {
 	}, {
 		.sect_name = "databases",
 		.set_key = parse_database,
+	}, {
+		.sect_name = "users",
+		.set_key = parse_user,
 	}, {
 		.sect_name = NULL,
 	}
@@ -302,16 +363,25 @@ void load_config(void)
 	if (ok) {
 		/* load users if needed */
 		if (cf_auth_type >= AUTH_TRUST)
-			load_auth_file(cf_auth_file);
+			loader_users_check();
 		loaded = true;
 	} else if (!loaded) {
 		die("Cannot load config file");
 	} else {
 		log_warning("Config file loading failed");
-		/* if ini file missing, dont kill anybody */
+		/* if ini file missing, don't kill anybody */
 		set_dbs_dead(false);
 	}
 
+	if (cf_auth_type == AUTH_HBA) {
+		struct HBA *hba = hba_load_rules(cf_auth_hba_file);
+		if (hba) {
+			if (parsed_hba)
+				hba_free(parsed_hba);
+			parsed_hba = hba;
+		}
+	}
+
 	/* reset pool_size, kill dbs */
 	config_postprocess();
 
@@ -445,7 +515,7 @@ static void go_daemon(void)
 	if (!cf_pidfile[0])
 		fatal("daemon needs pidfile configured");
 
-	/* dont log to stdout anymore */
+	/* don't log to stdout anymore */
 	cf_quiet = 1;
 
 	/* send stdin, stdout, stderr to /dev/null */
@@ -546,7 +616,7 @@ static void write_pidfile(void)
 		return;
 
 	pid = getpid();
-	snprintf(buf, sizeof(buf), "%u", (unsigned)pid);
+	snprintf(buf, sizeof(buf), "%u\n", (unsigned)pid);
 
 	fd = open(cf_pidfile, O_WRONLY | O_CREAT | O_EXCL, 0644);
 	if (fd < 0)
@@ -635,6 +705,9 @@ static void main_loop_once(void)
 	reuse_just_freed_objects();
 	rescue_timers();
 	per_loop_pooler_maint();
+
+	if (adns)
+		adns_per_loop(adns);
 }
 
 static void takeover_part1(void)
@@ -720,6 +793,8 @@ int main(int argc, char *argv[])
 	init_caches();
 	logging_prefix_cb = log_socket_prefix;
 
+	sbuf_tls_setup();
+
 	/* prefer cmdline over config for username */
 	if (arg_username) {
 		if (cf_username)
@@ -767,10 +842,11 @@ int main(int argc, char *argv[])
 	janitor_setup();
 	stats_setup();
 
-	if (did_takeover)
+	if (did_takeover) {
 		takeover_finish();
-	else
+	} else {
 		pooler_setup();
+	}
 
 	write_pidfile();
 
