@@ -94,6 +94,13 @@ static int user_node_cmp(uintptr_t userptr, struct AANode *node)
 	return strcmp(name, user->name);
 }
 
+/* destroy PgUser, for usage with btree */
+static void user_node_release(struct AANode *node, void *arg)
+{
+	PgUser *user = container_of(node, PgUser, tree_node);
+	slab_free(user_cache, user);
+}
+
 /* initialization before config loading */
 void init_objects(void)
 {
@@ -133,8 +140,13 @@ void change_client_state(PgSocket *client, SocketState newstate)
 		statlist_remove(&justfree_client_list, &client->head);
 		break;
 	case CL_LOGIN:
+		if (newstate == CL_WAITING)
+			newstate = CL_WAITING_LOGIN;
 		statlist_remove(&login_client_list, &client->head);
 		break;
+	case CL_WAITING_LOGIN:
+		if (newstate == CL_ACTIVE)
+			newstate = CL_LOGIN;
 	case CL_WAITING:
 		statlist_remove(&pool->waiting_client_list, &client->head);
 		break;
@@ -163,6 +175,7 @@ void change_client_state(PgSocket *client, SocketState newstate)
 		statlist_append(&login_client_list, &client->head);
 		break;
 	case CL_WAITING:
+	case CL_WAITING_LOGIN:
 		statlist_append(&pool->waiting_client_list, &client->head);
 		break;
 	case CL_ACTIVE:
@@ -229,12 +242,13 @@ void change_server_state(PgSocket *server, SocketState newstate)
 		statlist_append(&pool->tested_server_list, &server->head);
 		break;
 	case SV_IDLE:
-		if (server->close_needed || cf_server_round_robin)
+		if (server->close_needed || cf_server_round_robin) {
 			/* try to avoid immediate usage then */
 			statlist_append(&pool->idle_server_list, &server->head);
-		else
+		} else {
 			/* otherwise use LIFO */
 			statlist_prepend(&pool->idle_server_list, &server->head);
+		}
 		break;
 	case SV_ACTIVE:
 		statlist_append(&pool->active_server_list, &server->head);
@@ -281,9 +295,9 @@ static void put_in_order(struct List *newitem, struct StatList *list,
 
 	statlist_for_each(item, list) {
 		res = cmpfn(item, newitem);
-		if (res == 0)
+		if (res == 0) {
 			fatal("put_in_order: found existing elem");
-		else if (res > 0) {
+		} else if (res > 0) {
 			statlist_put_before(list, newitem, item);
 			return;
 		}
@@ -308,6 +322,7 @@ PgDatabase *add_database(const char *name)
 			slab_free(db_cache, db);
 			return NULL;
 		}
+		aatree_init(&db->user_tree, user_node_cmp, user_node_release);
 		put_in_order(&db->head, &database_list, cmp_database);
 	}
 
@@ -361,6 +376,32 @@ PgUser *add_user(const char *name, const char *passwd)
 		put_in_order(&user->head, &user_list, cmp_user);
 
 		aatree_insert(&user_tree, (uintptr_t)user->name, &user->tree_node);
+		user->pool_mode = POOL_INHERIT;
+	}
+	safe_strcpy(user->passwd, passwd, sizeof(user->passwd));
+	return user;
+}
+
+/* add or update db users */
+PgUser *add_db_user(PgDatabase *db, const char *name, const char *passwd)
+{
+	PgUser *user = NULL;
+	struct AANode *node;
+
+	node = aatree_search(&db->user_tree, (uintptr_t)name);
+	user = node ? container_of(node, PgUser, tree_node) : NULL;
+
+	if (user == NULL) {
+		user = slab_alloc(user_cache);
+		if (!user)
+			return NULL;
+
+		list_init(&user->head);
+		list_init(&user->pool_list);
+		safe_strcpy(user->name, name, sizeof(user->name));
+
+		aatree_insert(&db->user_tree, (uintptr_t)user->name, &user->tree_node);
+		user->pool_mode = POOL_INHERIT;
 	}
 	safe_strcpy(user->passwd, passwd, sizeof(user->passwd));
 	return user;
@@ -376,6 +417,7 @@ PgUser *force_user(PgDatabase *db, const char *name, const char *passwd)
 			return NULL;
 		list_init(&user->head);
 		list_init(&user->pool_list);
+		user->pool_mode = POOL_INHERIT;
 	}
 	safe_strcpy(user->name, name, sizeof(user->name));
 	safe_strcpy(user->passwd, passwd, sizeof(user->passwd));
@@ -470,7 +512,7 @@ PgPool *get_pool(PgDatabase *db, PgUser *user)
 /* deactivate socket and put into wait queue */
 static void pause_client(PgSocket *client)
 {
-	Assert(client->state == CL_ACTIVE);
+	Assert(client->state == CL_ACTIVE || client->state == CL_LOGIN);
 
 	slog_debug(client, "pause_client");
 	change_client_state(client, CL_WAITING);
@@ -481,7 +523,7 @@ static void pause_client(PgSocket *client)
 /* wake client from wait */
 void activate_client(PgSocket *client)
 {
-	Assert(client->state == CL_WAITING);
+	Assert(client->state == CL_WAITING || client->state == CL_WAITING_LOGIN);
 
 	slog_debug(client, "activate_client");
 	change_client_state(client, CL_ACTIVE);
@@ -515,7 +557,7 @@ bool check_fast_fail(PgSocket *client)
 		return true;
 	disconnect_client(client, true, "pgbouncer cannot connect to server");
 
-	/* usual relaunch wont work, as there are no waiting clients */
+	/* usual relaunch won't work, as there are no waiting clients */
 	launch_new_connection(pool);
 
 	return false;
@@ -529,25 +571,26 @@ bool find_server(PgSocket *client)
 	bool res;
 	bool varchange = false;
 
-	Assert(client->state == CL_ACTIVE);
+	Assert(client->state == CL_ACTIVE || client->state == CL_LOGIN);
 
 	if (client->link)
 		return true;
 
 	/* try to get idle server, if allowed */
-	if (cf_pause_mode == P_PAUSE) {
+	if (cf_pause_mode == P_PAUSE || pool->db->db_paused) {
 		server = NULL;
 	} else {
 		while (1) {
 			server = first_socket(&pool->idle_server_list);
-			if (!server)
+			if (!server) {
 				break;
-			else if (server->close_needed)
+			} else if (server->close_needed) {
 				disconnect_server(server, true, "obsolete connection");
-			else if (!server->ready)
+			} else if (!server->ready) {
 				disconnect_server(server, true, "idle server got dirty");
-			else
+			} else {
 				break;
+			}
 		}
 
 		if (!server && !check_fast_fail(client))
@@ -576,8 +619,9 @@ bool find_server(PgSocket *client)
 			res = false; /* don't process client data yet */
 			if (!sbuf_pause(&client->sbuf))
 				disconnect_client(client, true, "pause failed");
-		} else
+		} else {
 			res = true;
+		}
 	} else {
 		pause_client(client);
 		res = false;
@@ -596,8 +640,8 @@ static bool reuse_on_release(PgSocket *server)
 
 		/*
 		 * As the activate_client() does full read loop,
-		 * then it may happen that linked client close
-		 * couses server close.  Report it.
+		 * then it may happen that linked client closing
+		 * causes server closing.  Report it.
 		 */
 		if (server->state == SV_FREE || server->state == SV_JUSTFREE)
 			res = false;
@@ -653,16 +697,19 @@ bool release_server(PgSocket *server)
 		server->link->link = NULL;
 		server->link = NULL;
 
-		if (*cf_server_reset_query)
+		if (*cf_server_reset_query && (cf_server_reset_query_always ||
+					       pool_pool_mode(pool) == POOL_SESSION))
+		{
 			/* notify reset is required */
 			newstate = SV_TESTED;
-		else if (cf_server_check_delay == 0 && *cf_server_check_query)
+		} else if (cf_server_check_delay == 0 && *cf_server_check_query) {
 			/*
 			 * deprecated: before reset_query, the check_delay = 0
 			 * was used to get same effect.  This if() can be removed
 			 * after couple of releases.
 			 */
 			newstate = SV_USED;
+		}
 	case SV_USED:
 	case SV_TESTED:
 		break;
@@ -690,11 +737,12 @@ bool release_server(PgSocket *server)
 	slog_noise(server, "release_server: new state=%d", newstate);
 	change_server_state(server, newstate);
 
-	if (newstate == SV_IDLE)
+	if (newstate == SV_IDLE) {
 		/* immediately process waiters, to give fair chance */
 		return reuse_on_release(server);
-	else if (newstate == SV_TESTED)
+	} else if (newstate == SV_TESTED) {
 		return reset_on_release(server);
+	}
 
 	return true;
 }
@@ -760,6 +808,9 @@ void disconnect_server(PgSocket *server, bool notify, const char *reason, ...)
 		server->dns_token = NULL;
 	}
 
+	server->pool->db->connection_count--;
+	server->pool->user->connection_count--;
+
 	change_server_state(server, SV_JUSTFREE);
 	if (!sbuf_close(&server->sbuf))
 		log_noise("sbuf_close failed, retry later");
@@ -783,6 +834,7 @@ void disconnect_client(PgSocket *client, bool notify, const char *reason, ...)
 
 	switch (client->state) {
 	case CL_ACTIVE:
+	case CL_LOGIN:
 		if (client->link) {
 			PgSocket *server = client->link;
 			/* ->ready may be set before all is sent */
@@ -795,8 +847,8 @@ void disconnect_client(PgSocket *client, bool notify, const char *reason, ...)
 				disconnect_server(server, true, "unclean server");
 			}
 		}
-	case CL_LOGIN:
 	case CL_WAITING:
+	case CL_WAITING_LOGIN:
 	case CL_CANCEL:
 		break;
 	default:
@@ -833,8 +885,7 @@ static void connect_server(struct PgSocket *server, const struct sockaddr *sa, i
 		pga_copy(&server->remote_addr, sa);
 	}
 
-	if (cf_log_connections)
-		slog_info(server, "new connection to server");
+	slog_debug(server, "launching new connection to server");
 
 	/* start connecting */
 	res = sbuf_connect(&server->sbuf, sa, salen,
@@ -889,6 +940,7 @@ static void dns_connect(struct PgSocket *server)
 	const char *host = db->host;
 	const char *unix_dir;
 	int sa_len;
+	int res;
 
 	if (!host || host[0] == '/') {
 		slog_noise(server, "unix socket: %s", sa_un.sun_path);
@@ -904,23 +956,27 @@ static void dns_connect(struct PgSocket *server)
 			 "%s/.s.PGSQL.%d", unix_dir, db->port);
 		sa = (struct sockaddr *)&sa_un;
 		sa_len = sizeof(sa_un);
-	} else if (strchr(host, ':')) {  // assume IPv6 address on any : in addr
+		res = 1;
+	} else if (strchr(host, ':')) {  /* assume IPv6 address on any : in addr */
 		slog_noise(server, "inet6 socket: %s", db->host);
 		memset(&sa_in6, 0, sizeof(sa_in6));
 		sa_in6.sin6_family = AF_INET6;
-		inet_pton(AF_INET6, db->host, (void *) sa_in6.sin6_addr.s6_addr);
+		res = inet_pton(AF_INET6, db->host, &sa_in6.sin6_addr);
 		sa_in6.sin6_port = htons(db->port);
 		sa = (struct sockaddr *)&sa_in6;
 		sa_len = sizeof(sa_in6);
-	} else if (host[0] >= '0' && host[0] <= '9') { // else try IPv4
+	} else { /* else try IPv4 */
 		slog_noise(server, "inet socket: %s", db->host);
 		memset(&sa_in, 0, sizeof(sa_in));
 		sa_in.sin_family = AF_INET;
-		sa_in.sin_addr.s_addr = inet_addr(db->host);
+		res = inet_pton(AF_INET, db->host, &sa_in.sin_addr);
 		sa_in.sin_port = htons(db->port);
 		sa = (struct sockaddr *)&sa_in;
 		sa_len = sizeof(sa_in);
-	} else {
+	}
+
+	/* if simple parse failed, use DNS */
+	if (res != 1) {
 		struct DNSToken *tk;
 		slog_noise(server, "dns socket: %s", db->host);
 		/* launch dns lookup */
@@ -933,6 +989,67 @@ static void dns_connect(struct PgSocket *server)
 	connect_server(server, sa, sa_len);
 }
 
+PgSocket *compare_connections_by_time(PgSocket *lhs, PgSocket *rhs)
+{
+	if (!lhs)
+		return rhs;
+	if (!rhs)
+		return lhs;
+	return lhs->request_time < rhs->request_time ? lhs : rhs;
+}
+
+/* evict the single most idle connection from among all pools to make room in the db */
+bool evict_connection(PgDatabase *db)
+{
+	struct List *item;
+	PgPool *pool;
+	PgSocket *oldest_connection = NULL;
+
+	statlist_for_each(item, &pool_list) {
+		pool = container_of(item, PgPool, head);
+		if (pool->db != db)
+			continue;
+		oldest_connection = compare_connections_by_time(oldest_connection, last_socket(&pool->idle_server_list));
+		/* only evict testing connections if nobody's waiting */
+		if (statlist_empty(&pool->waiting_client_list)) {
+			oldest_connection = compare_connections_by_time(oldest_connection, last_socket(&pool->used_server_list));
+			oldest_connection = compare_connections_by_time(oldest_connection, last_socket(&pool->tested_server_list));
+		}
+	}
+
+	if (oldest_connection) {
+		disconnect_server(oldest_connection, true, "evicted");
+		return true;
+	}
+	return false;
+}
+
+/* evict the single most idle connection from among all pools to make room in the user */
+bool evict_user_connection(PgUser *user)
+{
+	struct List *item;
+	PgPool *pool;
+	PgSocket *oldest_connection = NULL;
+
+	statlist_for_each(item, &pool_list) {
+		pool = container_of(item, PgPool, head);
+		if (pool->user != user)
+			continue;
+		oldest_connection = compare_connections_by_time(oldest_connection, last_socket(&pool->idle_server_list));
+		/* only evict testing connections if nobody's waiting */
+		if (statlist_empty(&pool->waiting_client_list)) {
+			oldest_connection = compare_connections_by_time(oldest_connection, last_socket(&pool->used_server_list));
+			oldest_connection = compare_connections_by_time(oldest_connection, last_socket(&pool->tested_server_list));
+		}
+	}
+
+	if (oldest_connection) {
+		disconnect_server(oldest_connection, true, "evicted");
+		return true;
+	}
+	return false;
+}
+
 /* the pool needs new connection, if possible */
 void launch_new_connection(PgPool *pool)
 {
@@ -974,6 +1091,36 @@ void launch_new_connection(PgPool *pool)
 	}
 
 allow_new:
+	total = database_max_connections(pool->db);
+	if (total > 0) {
+		/* try to evict unused connections first */
+		while (pool->db->connection_count >= total) {
+			if (!evict_connection(pool->db)) {
+				break;
+			}
+		}
+		if (pool->db->connection_count >= total) {
+			log_debug("launch_new_connection: database full (%d >= %d)",
+					pool->db->connection_count, total);
+			return;
+		}
+	}
+
+	total = user_max_connections(pool->user);
+	if (total > 0) {
+		/* try to evict unused connection first */
+		while (pool->user->connection_count >= total) {
+			if (!evict_user_connection(pool->user)) {
+				break;
+			}
+		}
+		if (pool->user->connection_count >= total) {
+			log_debug("launch_new_connection: user full (%d >= %d)",
+					pool->user->connection_count, total);
+			return;
+		}
+	}
+
 	/* get free conn object */
 	server = slab_alloc(server_cache);
 	if (!server) {
@@ -987,6 +1134,8 @@ allow_new:
 	server->connect_time = get_cached_time();
 	pool->last_connect_time = get_cached_time();
 	change_server_state(server, SV_LOGIN);
+	pool->db->connection_count++;
+	pool->user->connection_count++;
 
 	dns_connect(server);
 }
@@ -1076,6 +1225,13 @@ void accept_cancel_request(PgSocket *req)
 				goto found;
 			}
 		}
+		statlist_for_each(citem, &pool->waiting_client_list) {
+			client = container_of(citem, PgSocket, head);
+			if (memcmp(client->cancel_key, req->cancel_key, 8) == 0) {
+				main_client = client;
+				goto found;
+			}
+		}
 	}
 found:
 
@@ -1130,6 +1286,8 @@ void forward_cancel_request(PgSocket *server)
 	Assert(server->state == SV_LOGIN);
 
 	SEND_CancelRequest(res, server, req->cancel_key);
+	if (!res)
+		log_warning("sending cancel request failed: %s", strerror(errno));
 
 	change_client_state(req, CL_JUSTFREE);
 }
@@ -1138,7 +1296,8 @@ bool use_client_socket(int fd, PgAddr *addr,
 		       const char *dbname, const char *username,
 		       uint64_t ckey, int oldfd, int linkfd,
 		       const char *client_enc, const char *std_string,
-		       const char *datestyle, const char *timezone)
+		       const char *datestyle, const char *timezone,
+		       const char *password)
 {
 	PgSocket *client;
 	PktBuf tmp;
@@ -1148,7 +1307,7 @@ bool use_client_socket(int fd, PgAddr *addr,
 		return false;
 	client->suspended = 1;
 
-	if (!set_pool(client, dbname, username))
+	if (!set_pool(client, dbname, username, password, true))
 		return false;
 
 	change_client_state(client, CL_ACTIVE);
@@ -1173,7 +1332,8 @@ bool use_server_socket(int fd, PgAddr *addr,
 		       const char *dbname, const char *username,
 		       uint64_t ckey, int oldfd, int linkfd,
 		       const char *client_enc, const char *std_string,
-		       const char *datestyle, const char *timezone)
+		       const char *datestyle, const char *timezone,
+		       const char *password)
 {
 	PgDatabase *db = find_database(dbname);
 	PgUser *user;
@@ -1189,10 +1349,13 @@ bool use_server_socket(int fd, PgAddr *addr,
 			return true;
 	}
 
-	if (db->forced_user)
+	if (db->forced_user) {
 		user = db->forced_user;
-	else
+	} else {
 		user = find_user(username);
+	}
+	if (!user && db->auth_user)
+		user = add_db_user(db, username, password);
 
 	pool = get_pool(db, user);
 	if (!pool)
@@ -1206,6 +1369,8 @@ bool use_server_socket(int fd, PgAddr *addr,
 	if (!res)
 		return false;
 
+	db->connection_count++;
+
 	server->suspended = 1;
 	server->pool = pool;
 	server->auth_user = user;
@@ -1337,9 +1502,26 @@ void tag_database_dirty(PgDatabase *db)
 
 void tag_autodb_dirty(void)
 {
-	struct List *item;
+	struct List *item, *tmp;
+	PgDatabase *db;
 	PgPool *pool;
 
+	/*
+	 * reload databases.
+	 */
+	statlist_for_each(item, &database_list) {
+		db = container_of(item, PgDatabase, head);
+		if (db->db_auto)
+			register_auto_database(db->name);
+	}
+	statlist_for_each_safe(item, &autodatabase_idle_list, tmp) {
+		db = container_of(item, PgDatabase, head);
+		if (db->db_auto)
+			register_auto_database(db->name);
+	}
+	/*
+	 * reload pools
+	 */
 	statlist_for_each(item, &pool_list) {
 		pool = container_of(item, PgPool, head);
 		if (pool->db->db_auto)
@@ -1387,17 +1569,19 @@ void reuse_just_freed_objects(void)
 
 	statlist_for_each_safe(item, &justfree_client_list, tmp) {
 		sk = container_of(item, PgSocket, head);
-		if (sbuf_is_closed(&sk->sbuf))
+		if (sbuf_is_closed(&sk->sbuf)) {
 			change_client_state(sk, CL_FREE);
-		else if (close_works)
+		} else if (close_works) {
 			close_works = sbuf_close(&sk->sbuf);
+		}
 	}
 	statlist_for_each_safe(item, &justfree_server_list, tmp) {
 		sk = container_of(item, PgSocket, head);
-		if (sbuf_is_closed(&sk->sbuf))
+		if (sbuf_is_closed(&sk->sbuf)) {
 			change_server_state(sk, SV_FREE);
-		else if (close_works)
+		} else if (close_works) {
 			close_works = sbuf_close(&sk->sbuf);
+		}
 	}
 }
 
