@@ -87,14 +87,13 @@ struct PktBuf *pktbuf_temp(void)
 
 bool pktbuf_send_immediate(PktBuf *buf, PgSocket *sk)
 {
-	int fd = sbuf_socket(&sk->sbuf);
 	uint8_t *pos = buf->buf + buf->send_pos;
 	int amount = buf->write_pos - buf->send_pos;
 	int res;
 
 	if (buf->failed)
 		return false;
-	res = safe_send(fd, pos, amount, 0);
+	res = sbuf_op_send(&sk->sbuf, pos, amount);
 	if (res < 0) {
 		log_debug("pktbuf_send_immediate: %s", strerror(errno));
 	}
@@ -104,6 +103,7 @@ bool pktbuf_send_immediate(PktBuf *buf, PgSocket *sk)
 static void pktbuf_send_func(int fd, short flags, void *arg)
 {
 	PktBuf *buf = arg;
+	SBuf *sbuf = &buf->queued_dst->sbuf;
 	int amount, res;
 
 	log_debug("pktbuf_send_func(%d, %d, %p)", fd, (int)flags, buf);
@@ -112,9 +112,9 @@ static void pktbuf_send_func(int fd, short flags, void *arg)
 		return;
 
 	amount = buf->write_pos - buf->send_pos;
-	res = safe_send(fd, buf->buf + buf->send_pos, amount, 0);
+	res = sbuf_op_send(sbuf, buf->buf + buf->send_pos, amount);
 	if (res < 0) {
-		if (res == EAGAIN) {
+		if (errno == EAGAIN) {
 			res = 0;
 		} else {
 			log_error("pktbuf_send_func: %s", strerror(errno));
@@ -131,14 +131,13 @@ static void pktbuf_send_func(int fd, short flags, void *arg)
 			log_error("pktbuf_send_func: %s", strerror(errno));
 			pktbuf_free(buf);
 		}
-	} else
+	} else {
 		pktbuf_free(buf);
+	}
 }
 
 bool pktbuf_send_queued(PktBuf *buf, PgSocket *sk)
 {
-	int fd = sbuf_socket(&sk->sbuf);
-
 	Assert(!buf->sending);
 	Assert(!buf->fixed_buf);
 
@@ -147,7 +146,8 @@ bool pktbuf_send_queued(PktBuf *buf, PgSocket *sk)
 		return send_pooler_error(sk, true, "result prepare failed");
 	} else {
 		buf->sending = 1;
-		pktbuf_send_func(fd, EV_WRITE, buf);
+		buf->queued_dst = sk;
+		pktbuf_send_func(sk->sbuf.sock, EV_WRITE, buf);
 		return true;
 	}
 }
@@ -369,8 +369,9 @@ void pktbuf_write_RowDescription(PktBuf *buf, const char *tupdesc, ...)
 		} else if (tupdesc[i] == 'T') {
 			pktbuf_put_uint32(buf, TEXTOID);
 			pktbuf_put_uint16(buf, -1);
-		} else
+		} else {
 			fatal("bad tupdesc");
+		}
 		pktbuf_put_uint32(buf, 0);
 		pktbuf_put_uint16(buf, 0);
 	}
@@ -412,8 +413,9 @@ void pktbuf_write_DataRow(PktBuf *buf, const char *tupdesc, ...)
 		} else if (tupdesc[i] == 'T') {
 			usec_t time = va_arg(ap, usec_t);
 			val = format_time_s(time, tmp, sizeof(tmp));
-		} else
+		} else {
 			fatal("bad tupdesc: %s", tupdesc);
+		}
 
 		if (val) {
 			len = strlen(val);
@@ -429,3 +431,44 @@ void pktbuf_write_DataRow(PktBuf *buf, const char *tupdesc, ...)
 	pktbuf_finish_packet(buf);
 }
 
+/*
+ * Send Parse+Bind+Execute with string parameters.
+ */
+void pktbuf_write_ExtQuery(PktBuf *buf, const char *query, int nargs, ...)
+{
+	va_list ap;
+	const char *val;
+	int len, i;
+
+	/* Parse */
+	pktbuf_write_generic(buf, 'P', "csh", 0, query, 0);
+
+	/* Bind */
+	pktbuf_start_packet(buf, 'B');
+	pktbuf_put_char(buf, 0);	/* portal name */
+	pktbuf_put_char(buf, 0);	/* query name */
+	pktbuf_put_uint16(buf, 0);	/* number of parameter format codes */
+	pktbuf_put_uint16(buf, nargs);	/* number of parameter values */
+
+	va_start(ap, nargs);
+	for (i = 0; i < nargs; i++) {
+		val = va_arg(ap, char *);
+		len = strlen(val);
+		pktbuf_put_uint32(buf, len);
+		pktbuf_put_bytes(buf, val, len);
+	}
+	va_end(ap);
+
+	pktbuf_put_uint16(buf, 0);	/* number of result-column format codes */
+	pktbuf_finish_packet(buf);
+
+	/* Describe */
+	pktbuf_write_generic(buf, 'D', "cc", 'P', 0);
+
+	/* Execute */
+	pktbuf_write_generic(buf, 'E', "ci", 0, 0);
+
+	/* Sync */
+	pktbuf_write_generic(buf, 'S', "");
+}
+
