@@ -141,7 +141,7 @@ static bool handle_server_startup(PgSocket *server, PktHdr *pkt)
 			/* deliberately ignore transaction status */
 		} else if (server->pool->db->connect_query) {
 			server->exec_on_connect = 1;
-			slog_debug(server, "server conect ok, send exec_on_connect");
+			slog_debug(server, "server connect ok, send exec_on_connect");
 			SEND_generic(res, server, 'Q', "s", server->pool->db->connect_query);
 			if (!res)
 				disconnect_server(server, false, "exec_on_connect query failed");
@@ -185,6 +185,34 @@ static bool handle_server_startup(PgSocket *server, PktHdr *pkt)
 	return res;
 }
 
+int pool_pool_mode(PgPool *pool)
+{
+	int pool_mode = pool->user->pool_mode;
+	if (pool_mode == POOL_INHERIT)
+		pool_mode = pool->db->pool_mode;
+	if (pool_mode == POOL_INHERIT)
+		pool_mode = cf_pool_mode;
+	return pool_mode;
+}
+
+int database_max_connections(PgDatabase *db)
+{
+	if (db->max_db_connections <= 0) {
+		return cf_max_db_connections;
+        } else {
+		return db->max_db_connections;
+	}
+}
+
+int user_max_connections(PgUser *user)
+{
+	if (user->max_user_connections <= 0) {
+		return cf_max_user_connections;
+	} else {
+		return user->max_user_connections;
+	}
+}
+
 /* process packets on logged in connection */
 static bool handle_server_work(PgSocket *server, PktHdr *pkt)
 {
@@ -210,14 +238,22 @@ static bool handle_server_work(PgSocket *server, PktHdr *pkt)
 			return false;
 
 		/* set ready only if no tx */
-		if (state == 'I') {
+		if (state == 'I')
 			ready = true;
-		} else if (cf_pool_mode == POOL_STMT) {
+		else if (pool_pool_mode(server->pool) == POOL_STMT) {
 			disconnect_server(server, true, "Long transactions not allowed");
 			return false;
 		} else if (state == 'T' || state == 'E') {
 			idle_tx = true;
 		}
+
+		if (client && !server->setting_vars) {
+			if (client->expect_rfq_count > 0) {
+				client->expect_rfq_count--;
+			} else if (server->state == SV_ACTIVE) {
+				slog_debug(client, "unexpected ReadyForQuery - expect_rfq_count=%d", client->expect_rfq_count);
+			}
+		}
 		break;
 
 	case 'S':		/* ParameterStatus */
@@ -227,12 +263,12 @@ static bool handle_server_work(PgSocket *server, PktHdr *pkt)
 
 	/*
 	 * 'E' and 'N' packets currently set ->ready to 0.  Correct would
-	 * be to leave ->ready as-is, because overal TX state stays same.
+	 * be to leave ->ready as-is, because overall TX state stays same.
 	 * It matters for connections in IDLE or USED state which get dirty
 	 * suddenly but should not as they are still usable.
 	 *
 	 * But the 'E' or 'N' packet between transactions signifies probably
-	 * dying backend.  This its better to tag server as dirty and drop
+	 * dying backend.  It is better to tag server as dirty and drop
 	 * it later.
 	 */
 	case 'E':		/* ErrorResponse */
@@ -251,6 +287,17 @@ static bool handle_server_work(PgSocket *server, PktHdr *pkt)
 			disconnect_server(server, true, "invalid server parameter");
 			return false;
 		}
+	case 'C':		/* CommandComplete */
+
+		/* ErrorResponse and CommandComplete show end of copy mode */
+		if (server->copy_mode) {
+			server->copy_mode = false;
+
+			/* it's impossible to track sync count over copy */
+			if (client)
+				client->expect_rfq_count = 0;
+		}
+		break;
 
 	case 'N':		/* NoticeResponse */
 		break;
@@ -261,6 +308,11 @@ static bool handle_server_work(PgSocket *server, PktHdr *pkt)
 		ready = server->ready;
 		break;
 
+	/* copy mode */
+	case 'G':		/* CopyInResponse */
+	case 'H':		/* CopyOutResponse */
+		server->copy_mode = true;
+		break;
 	/* chat packets */
 	case '2':		/* BindComplete */
 	case '3':		/* CloseComplete */
@@ -269,11 +321,8 @@ static bool handle_server_work(PgSocket *server, PktHdr *pkt)
 	case 'I':		/* EmptyQueryResponse == CommandComplete */
 	case 'V':		/* FunctionCallResponse */
 	case 'n':		/* NoData */
-	case 'G':		/* CopyInResponse */
-	case 'H':		/* CopyOutResponse */
 	case '1':		/* ParseComplete */
 	case 's':		/* PortalSuspended */
-	case 'C':		/* CommandComplete */
 
 	/* data packets, there will be more coming */
 	case 'd':		/* CopyData(F/B) */
@@ -290,15 +339,19 @@ static bool handle_server_work(PgSocket *server, PktHdr *pkt)
 		Assert(client);
 		sbuf_prepare_skip(sbuf, pkt->len);
 	} else if (client) {
-		sbuf_prepare_send(sbuf, &client->sbuf, pkt->len);
-		if (ready && client->query_start) {
-			usec_t total;
-			total = get_cached_time() - client->query_start;
-			client->query_start = 0;
-			server->pool->stats.query_time += total;
-			slog_debug(client, "query time: %d us", (int)total);
-		} else if (ready) {
-			slog_warning(client, "FIXME: query end, but query_start == 0");
+		if (client->state == CL_LOGIN) {
+			return handle_auth_response(client, pkt);
+		} else {
+			sbuf_prepare_send(sbuf, &client->sbuf, pkt->len);
+			if (ready && client->query_start) {
+				usec_t total;
+				total = get_cached_time() - client->query_start;
+				client->query_start = 0;
+				server->pool->stats.query_time += total;
+				slog_debug(client, "query time: %d us", (int)total);
+			} else if (ready) {
+				slog_warning(client, "FIXME: query end, but query_start == 0");
+			}
 		}
 	} else {
 		if (server->state != SV_TESTED)
@@ -316,9 +369,18 @@ static bool handle_connect(PgSocket *server)
 {
 	bool res = false;
 	PgPool *pool = server->pool;
+	char buf[PGADDR_BUF + 32];
 
 	fill_local_addr(server, sbuf_socket(&server->sbuf), pga_is_unix(&server->remote_addr));
 
+	if (cf_log_connections) {
+		if (pga_is_unix(&server->remote_addr))
+			slog_info(server, "new connection to server");
+		else
+			slog_info(server, "new connection to server (from %s)",
+				  pga_str(&server->local_addr, buf, sizeof(buf)));
+	}
+
 	if (!statlist_empty(&pool->cancel_req_list)) {
 		slog_debug(server, "use it for pending cancel req");
 		/* if pending cancel req, send it */
@@ -328,13 +390,53 @@ static bool handle_connect(PgSocket *server)
 		disconnect_server(server, false, "sent cancel req");
 	} else {
 		/* proceed with login */
-		res = send_startup_packet(server);
+		if (cf_server_tls_sslmode > SSLMODE_DISABLED) {
+			slog_noise(server, "P: SSL request");
+			res = send_sslreq_packet(server);
+			if (res)
+				server->wait_sslchar = true;
+		} else {
+			slog_noise(server, "P: startup");
+			res = send_startup_packet(server);
+		}
 		if (!res)
 			disconnect_server(server, false, "startup pkt failed");
 	}
 	return res;
 }
 
+static bool handle_sslchar(PgSocket *server, struct MBuf *data)
+{
+	uint8_t schar = '?';
+	bool ok;
+
+	server->wait_sslchar = false;
+
+	ok = mbuf_get_byte(data, &schar);
+	if (!ok || (schar != 'S' && schar != 'N') || mbuf_avail_for_read(data) != 0) {
+		disconnect_server(server, false, "bad sslreq answer");
+		return false;
+	}
+
+	if (schar == 'S') {
+		slog_noise(server, "launching tls");
+		ok = sbuf_tls_connect(&server->sbuf, server->pool->db->host);
+	} else if (cf_server_tls_sslmode >= SSLMODE_REQUIRE) {
+		disconnect_server(server, false, "server refused SSL");
+		return false;
+	} else {
+		/* proceed with non-TLS connection */
+		ok = send_startup_packet(server);
+	}
+
+	if (ok) {
+		sbuf_prepare_skip(&server->sbuf, 1);
+	} else {
+		disconnect_server(server, false, "sslreq processing failed");
+	}
+	return ok;
+}
+
 /* callback from SBuf */
 bool server_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 {
@@ -342,6 +444,7 @@ bool server_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 	PgSocket *server = container_of(sbuf, PgSocket, sbuf);
 	PgPool *pool = server->pool;
 	PktHdr pkt;
+	char infobuf[96];
 
 	Assert(is_server_socket(server));
 	Assert(server->state != SV_FREE);
@@ -358,7 +461,11 @@ bool server_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 		disconnect_client(server->link, false, "unexpected eof");
 		break;
 	case SBUF_EV_READ:
-		if (mbuf_avail_for_read(data) < NEW_HEADER_LEN) {
+		if (server->wait_sslchar) {
+			res = handle_sslchar(server, data);
+			break;
+		}
+		if (incomplete_header(data)) {
 			slog_noise(server, "S: got partial header, trying to wait a bit");
 			break;
 		}
@@ -408,11 +515,18 @@ bool server_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 			sbuf_continue(&client->sbuf);
 			break;
 		}
-		
-		if (cf_pool_mode  != POOL_SESSION || server->state == SV_TESTED) {
+
+		if (pool_pool_mode(pool) != POOL_SESSION || server->state == SV_TESTED || server->resetting) {
+			server->resetting = false;
 			switch (server->state) {
 			case SV_ACTIVE:
 			case SV_TESTED:
+				/* keep link if client expects more Syncs */
+				if (server->link) {
+					if (server->link->expect_rfq_count > 0)
+						break;
+				}
+
 				/* retval does not matter here */
 				release_server(server);
 				break;
@@ -426,6 +540,23 @@ bool server_proto(SBuf *sbuf, SBufEvent evtype, struct MBuf *data)
 	case SBUF_EV_PKT_CALLBACK:
 		slog_warning(server, "SBUF_EV_PKT_CALLBACK with state=%d", server->state);
 		break;
+	case SBUF_EV_TLS_READY:
+		Assert(server->state == SV_LOGIN);
+
+		tls_get_connection_info(server->sbuf.tls, infobuf, sizeof infobuf);
+		if (cf_log_connections) {
+			slog_info(server, "SSL established: %s", infobuf);
+		} else {
+			slog_noise(server, "SSL established: %s", infobuf);
+		}
+
+		server->request_time = get_cached_time();
+		res = send_startup_packet(server);
+		if (res)
+			sbuf_continue(&server->sbuf);
+		else
+			disconnect_server(server, false, "TLS startup failed");
+		break;
 	}
 	if (!res && pool->db->admin)
 		takeover_login_failed();
