@@ -22,8 +22,8 @@
 
 #include "bouncer.h"
 
-#define MD5_COMPAT
-#include <usual/md5.h>
+#include <usual/crypto/md5.h>
+#include <usual/crypto/csrandom.h>
 
 int log_socket_prefix(enum LogLevel lev, void *ctx, char *dst, unsigned int dstlen)
 {
@@ -31,24 +31,36 @@ int log_socket_prefix(enum LogLevel lev, void *ctx, char *dst, unsigned int dstl
 	const char *user, *db, *host;
 	char host6[PGADDR_BUF];
 	int port;
+	char stype;
 
 	/* no prefix */
 	if (!sock)
 		return 0;
 
 	/* format prefix */
+	stype = is_server_socket(sock) ? 'S' : 'C';
+	port = pga_port(&sock->remote_addr);
 	db = sock->pool ? sock->pool->db->name : "(nodb)";
 	user = sock->auth_user ? sock->auth_user->name : "(nouser)";
 	if (pga_is_unix(&sock->remote_addr)) {
-		host = "unix";
+		unsigned long pid = sock->remote_addr.scred.pid;
+		if (pid) {
+			snprintf(host6, sizeof(host6), "unix(%lu)", pid);
+			host = host6;
+		} else {
+			host = "unix";
+		}
 	} else {
 		host = pga_ntop(&sock->remote_addr, host6, sizeof(host6));
 	}
-	port = pga_port(&sock->remote_addr);
 
-	return snprintf(dst, dstlen, "%c-%p: %s/%s@%s:%d ",
-			is_server_socket(sock) ? 'S' : 'C',
-			sock, db, user, host, port);
+	if (pga_family(&sock->remote_addr) == AF_INET6) {
+		return snprintf(dst, dstlen, "%c-%p: %s/%s@[%s]:%d ",
+			stype, sock, db, user, host, port);
+	} else {
+		return snprintf(dst, dstlen, "%c-%p: %s/%s@%s:%d ",
+			stype, sock, db, user, host, port);
+	}
 }
 
 const char *bin2hex(const uint8_t *src, unsigned srclen, char *dst, unsigned dstlen)
@@ -80,13 +92,13 @@ void pg_md5_encrypt(const char *part1,
 		    const char *part2, size_t part2len,
 		    char *dest)
 {
-	MD5_CTX ctx;
+	struct md5_ctx ctx;
 	uint8_t hash[MD5_DIGEST_LENGTH];
 
-	MD5_Init(&ctx);
-	MD5_Update(&ctx, part1, strlen(part1));
-	MD5_Update(&ctx, part2, part2len);
-	MD5_Final(hash, &ctx);
+	md5_reset(&ctx);
+	md5_update(&ctx, part1, strlen(part1));
+	md5_update(&ctx, part2, part2len);
+	md5_final(&ctx, hash);
 
 	memcpy(dest, "md5", 3);
 	hash2hex(hash, dest + 3);
@@ -95,9 +107,7 @@ void pg_md5_encrypt(const char *part1,
 /* wrapped for getting random bytes */
 void get_random_bytes(uint8_t *dest, int len)
 {
-	int i;
-	for (i = 0; i < len; i++)
-		dest[i] = random() & 255;
+	csrandom_bytes(dest, len);
 }
 
 /* set needed socket options */
@@ -105,71 +115,32 @@ bool tune_socket(int sock, bool is_unix)
 {
 	int res;
 	int val;
+	bool ok;
 
-	/* close fd on exec */
-	res = fcntl(sock, F_SETFD, FD_CLOEXEC);
-	if (res < 0)
+	/*
+	 * Generic stuff + nonblock.
+	 */
+	ok = socket_setup(sock, true);
+	if (!ok)
 		goto fail;
 
-	/* when no data available, return EAGAIN instead blocking */
-	socket_set_nonblocking(sock, 1);
-
-#ifdef SO_NOSIGPIPE
-	/* disallow SIGPIPE, if possible */
-	val = 1;
-	res = setsockopt(sock, SOL_SOCKET, SO_NOSIGPIPE, &val, sizeof(val));
-	if (res < 0)
-		goto fail;
-#endif
-
 	/*
 	 * Following options are for network sockets
 	 */
 	if (is_unix)
 		return true;
 
-	/* the keepalive stuff needs some poking before enbling */
-	if (cf_tcp_keepalive) {
-		/* turn on socket keepalive */
-		val = 1;
-		res = setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &val, sizeof(val));
-		if (res < 0)
-			goto fail;
-#ifdef __linux__
-		/* set count of keepalive packets */
-		if (cf_tcp_keepcnt > 0) {
-			val = cf_tcp_keepcnt;
-			res = setsockopt(sock, IPPROTO_TCP, TCP_KEEPCNT, &val, sizeof(val));
-			if (res < 0)
-				goto fail;
-		}
-		/* how long the connection can stay idle before sending keepalive pkts */
-		if (cf_tcp_keepidle) {
-			val = cf_tcp_keepidle;
-			res = setsockopt(sock, IPPROTO_TCP, TCP_KEEPIDLE, &val, sizeof(val));
-			if (res < 0)
-				goto fail;
-		}
-		/* time between packets */
-		if (cf_tcp_keepintvl) {
-			val = cf_tcp_keepintvl;
-			res = setsockopt(sock, IPPROTO_TCP, TCP_KEEPINTVL, &val, sizeof(val));
-			if (res < 0)
-				goto fail;
-		}
-#else
-#ifdef TCP_KEEPALIVE
-		if (cf_tcp_keepidle) {
-			val = cf_tcp_keepidle;
-			res = setsockopt(sock, IPPROTO_TCP, TCP_KEEPALIVE, &val, sizeof(val));
-			if (res < 0)
-				goto fail;
-		}
-#endif
-#endif
-	}
+	/*
+	 * TCP Keepalive
+	 */
+	ok = socket_set_keepalive(sock, cf_tcp_keepalive, cf_tcp_keepidle,
+				  cf_tcp_keepintvl, cf_tcp_keepcnt);
+	if (!ok)
+		goto fail;
 
-	/* set in-kernel socket buffer size */
+	/*
+	 * set in-kernel socket buffer size
+	 */
 	if (cf_tcp_socket_buffer) {
 		val = cf_tcp_socket_buffer;
 		res = setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &val, sizeof(val));
@@ -212,7 +183,7 @@ loop:
 
 	/* move listpos further */
 	listpos = p + len;
-	/* survive len=0 and avoid unneccesary compare */
+	/* survive len=0 and avoid unnecessary compare */
 	if (*listpos)
 		listpos++;
 
@@ -238,7 +209,17 @@ void fill_remote_addr(PgSocket *sk, int fd, bool is_unix)
 	int err;
 
 	if (is_unix) {
+		uid_t uid = 0;
+		gid_t gid = 0;
+		pid_t pid = 0;
 		pga_set(dst, AF_UNIX, cf_listen_port);
+		if (getpeercreds(fd, &uid, &gid, &pid) >= 0) {
+			log_noise("unix peer uid: %d", (int)uid);
+		} else {
+			log_warning("unix peer uid failed: %s", strerror(errno));
+		}
+		dst->scred.uid = uid;
+		dst->scred.pid = pid;
 	} else {
 		err = getpeername(fd, (struct sockaddr *)dst, &len);
 		if (err < 0) {
@@ -256,6 +237,8 @@ void fill_local_addr(PgSocket *sk, int fd, bool is_unix)
 
 	if (is_unix) {
 		pga_set(dst, AF_UNIX, cf_listen_port);
+		dst->scred.uid = geteuid();
+		dst->scred.pid = getpid();
 	} else {
 		err = getsockname(fd, (struct sockaddr *)dst, &len);
 		if (err < 0) {
@@ -350,11 +333,6 @@ void pga_copy(PgAddr *a, const struct sockaddr *sa)
 	}
 }
 
-static inline unsigned pga_family(const PgAddr *a)
-{
-	return a->sa.sa_family;
-}
-
 int pga_cmp_addr(const PgAddr *a, const PgAddr *b)
 {
     if (pga_family(a) != pga_family(b))
@@ -426,6 +404,42 @@ const char *pga_str(const PgAddr *a, char *dst, int dstlen)
 {
 	char buf[PGADDR_BUF];
 	pga_ntop(a, buf, sizeof(buf));
-	snprintf(dst, dstlen, "%s@%d", buf, pga_port(a));
+	if (pga_family(a) == AF_INET6) {
+		snprintf(dst, dstlen, "[%s]:%d", buf, pga_port(a));
+	} else if (pga_family(a) == AF_UNIX && a->scred.pid) {
+		snprintf(dst, dstlen, "%s:%d$%u", buf, pga_port(a), a->scred.pid);
+	} else {
+		snprintf(dst, dstlen, "%s:%d", buf, pga_port(a));
+	}
 	return dst;
 }
+
+static const char *cached_hostname(void)
+{
+	static char cache[256];
+	int err;
+
+	if (cache[0] == 0) {
+		err = gethostname(cache, sizeof(cache));
+		if (err != 0)
+			strlcpy(cache, "somehost", sizeof(cache));
+	}
+	return cache;
+}
+
+const char *pga_details(const PgAddr *a, char *dst, int dstlen)
+{
+	char buf[PGADDR_BUF];
+	pga_ntop(a, buf, sizeof(buf));
+	if (pga_family(a) == AF_INET6) {
+		snprintf(dst, dstlen, "[%s]:%d", buf, pga_port(a));
+	} else if (pga_family(a) == AF_UNIX && a->scred.pid) {
+		snprintf(dst, dstlen, "%s(%u@%s):%d", buf, a->scred.pid, cached_hostname(), pga_port(a));
+	} else {
+		snprintf(dst, dstlen, "%s:%d", buf, pga_port(a));
+	}
+	return dst;
+}
+
+
+
